import pytest
import numpy as np
import logging, re
from typing import Type

from cosapp.base import System, Port
from cosapp.core import MathematicalProblem
from cosapp.drivers import NonLinearSolver, RunSingleCase
from cosapp.utils.testing import get_args, DummySystemFactory


class AbcPort(Port):
    def setup(self):
        self.add_variable('a', 1.0)
        self.add_variable('b', 1.0)
        self.add_variable('c', np.zeros(4))


class XyzPort(Port):
    def setup(self):
        self.add_variable('x', 1.0)
        self.add_variable('y', 1.0)
        self.add_variable('z', 1.0)


@pytest.fixture
def A_factory() -> Type[System]:
    def factory(classname, **settings):
        return DummySystemFactory(
            classname,
            inputs = get_args(XyzPort, 'p_in'),
            outputs = get_args(AbcPort, 'q_out'),
            inwards = [
                get_args('x', 1.0),
                get_args('y', 0.5),
                get_args('u', np.ones(3)),
            ],
            outwards = [get_args('z', 0.0)],
            **settings,
        )
    return factory


@pytest.fixture
def B_factory() -> Type[System]:
    def factory(classname, **settings):
        return DummySystemFactory(
            classname,
            inwards = [
                get_args('length', 1.0),
                get_args('width', 0.5),
            ],
            outwards = [
                get_args('area', 0.0),
                get_args('v', np.zeros(4)),
            ],
            inputs = get_args(XyzPort, 'p_in'),
            **settings,
        )
    return factory


@pytest.fixture
def A(A_factory) -> Type[System]:
    return A_factory("A")


@pytest.fixture
def Aun(A_factory) -> Type[System]:
    return A_factory("Aun",
        unknowns = "p_in.x",
    )


@pytest.fixture
def Atrg(A_factory) -> Type[System]:
    return A_factory("Atrg",
        targets = [
            "z",
            "q_out.a",
            "abs(q_out.b)",
            "q_out.c[::2]",
        ],
    )


@pytest.fixture
def B(B_factory) -> Type[System]:
    return B_factory("B")


@pytest.fixture
def Beq(B_factory) -> Type[System]:
    return B_factory("Beq",
        equations = "v[::2] == [0, 1]",
    )


@pytest.fixture
def Bun(B_factory) -> Type[System]:
    return B_factory("Bun",
        unknowns = get_args(["p_in.x", "p_in.y"]),
    )


@pytest.fixture
def s1(A, Aun, B, Beq):
    """Generates composite system:

             top
         _____|_____
        |     |     |
        a     b    foo
        |           |
       aun         beq
    """
    a: System = A('a')
    a.add_child(Aun('aun'), pulling={'p_in': 'p'})
    top = System('s1')
    top.add_child(a, pulling=['x', 'y', 'q_out'])
    top.add_child(B('b'))
    foo = System('foo')
    foo.add_child(Beq('beq'), pulling={'v': 'v', 'p_in': 'xyz_in'})
    top.add_child(foo)
    return top


@pytest.fixture
def s2(Atrg):
    """Generates simple composite system `top/sub`.
    Sub-system `sub` has targets, some of which are pulled.
    """
    top = System('s2')
    top.add_child(Atrg('sub'), pulling={'q_out': 'out'})
    # Initialize outputs manually
    top.sub.z = 0.12
    top.out.set_values(
        a = 3.14,
        b = -2.3,
        c = np.r_[0.1, 0.2, 0.3, 0.4],
    )
    top.sub.q_out.set_values(
        a = 0.0,
        b = 0.0,
        c = np.zeros(4),
    )
    return top


def test_s1(s1):
    """Check fixture `s1` tree"""
    get_name = lambda system: system.name
    assert list(s1.children) == ['a', 'b', 'foo']
    assert list(map(get_name, s1.tree())) == [
        'aun', 'a', 'b', 'beq', 'foo', 's1',
    ]


def test_MathematicalProblem_repr_1(s1: System, caplog):
    """Test off-design unknown and equation assembly in system `s1`"""
    solver = s1.add_driver(NonLinearSolver('solver'))

    with caplog.at_level(logging.INFO):
        for driver in solver.tree():
            driver.setup_run()
    
    problem = solver.problem
    assert problem is not None
    assert isinstance(problem, MathematicalProblem)
    assert len(caplog.records) == 0
    assert set(problem.residues) == {
        "foo.beq: v[::2] == [0, 1]",
    }
    assert set(problem.unknowns) == {
        "a.p.x",
    }
    pattern = "\n".join([
        r"Unknowns \[1\]",
        r"  a\.p\.x = .*",
        r"Equations \[2\]",
        r"  foo\.beq: v\[::2\] == \[0, 1\] := \[.* .*\]",
    ])
    assert re.match(pattern, repr(problem))


def test_MathematicalProblem_repr_2(s1: System, caplog):
    """Same as `test_MathematicalProblem_keys_1`, with additional
    unknowns and equations added at solver level.
    """
    solver = s1.add_driver(NonLinearSolver('solver'))
    solver.add_unknown(['b.width', 'a.aun.p_in.y'])
    solver.add_equation("b.area == 10")

    with caplog.at_level(logging.INFO):
        for driver in solver.tree():
            driver.setup_run()
    
    problem = solver.problem
    assert problem is not None

    # print("", problem, sep="\n")
    assert len(caplog.records) > 0
    assert re.match(
        "Replace unknown 'a.aun.p_in.y' by 'a.p.y'",
        caplog.records[0].message
    )
    # Note: in a single-point problem, all equations
    # and off-design unknowns should be plainly displayed,
    # i.e. no wrapping of the kind "runner[b.area == 10]".
    assert set(problem.residues) == {
        "foo.beq: v[::2] == [0, 1]",
        "b.area == 10",
    }
    assert set(problem.unknowns) == {
        "a.p.x",
        "a.p.y",
        "b.width",
    }
    pattern = "\n".join([
        r"Unknowns \[3\]",
        r"  b\.width = .*",
        r"  a\.p\.y = .*",
        r"  a\.p\.x = .*",
        r"Equations \[3\]",
        r"  b\.area == 10 := .*",
        r"  foo\.beq: v\[::2\] == \[0, 1\] := \[.* .*\]",
    ])
    # print(pattern, problem, sep="\n")
    assert re.match(pattern, repr(problem))


def test_MathematicalProblem_repr_3(s1: System, caplog):
    """Check unknowns and equations in a multi-point solver.
    """
    solver = s1.add_driver(NonLinearSolver('solver'))
    point1 = solver.add_driver(RunSingleCase('point1'))
    point2 = solver.add_driver(RunSingleCase('point2'))

    solver.add_unknown(['b.width', 'b.length'])
    solver.add_equation("q_out.a == x + y")
    point1.add_unknown('a.aun.u[-1]').add_equation("a.z == 0")
    point2.design.add_unknown('foo.beq.p_in.y').add_equation("q_out.b == 0")

    with caplog.at_level(logging.INFO):
        for driver in solver.tree():
            driver.setup_run()
    
    assert len(caplog.records) > 0
    assert re.match(
        "Replace unknown 'foo.beq.p_in.y' by 'foo.xyz_in.y'",
        caplog.records[0].message
    )
    problem = solver.problem
    assert problem is not None
    # print("", problem, sep="\n")

    assert set(problem.residues) == {
        "point1[foo.beq: v[::2] == [0, 1]]",
        "point2[foo.beq: v[::2] == [0, 1]]",
        "point1[q_out.a == x + y]",
        "point2[q_out.a == x + y]",
        "point2[q_out.b == 0]",
        "point1[a.z == 0]",
    }
    assert set(problem.unknowns) == {
        # Design unknowns:
        "b.width",
        "b.length",
        "foo.xyz_in.y",
        # Off-design unknowns:
        "point1[a.p.x]",
        "point2[a.p.x]",
        "point1[a.aun.u[-1]]",
    }
    pattern = "\n".join([
        r"Unknowns \[6\]",
        r"  b\.width = .*",
        r"  b\.length = .*",
        r"  point1\[a\.aun\.u\[-1\]\] = \[.*\]",
        r"  point1\[a\.p\.x\] = .*",
        r"  foo\.xyz_in\.y = .*",
        r"  point2\[a\.p\.x\] = .*",
        r"Equations \[8\]",
        r"  point1\[a\.z == 0\] := .*",
        r"  point1\[foo\.beq: v\[::2\] == \[0, 1\]\] := \[.* .*\]",
        r"  point1\[q_out\.a == x \+ y\] := .*",
        r"  point2\[q_out\.b == 0\] := .*",
        r"  point2\[foo\.beq: v\[::2\] == \[0, 1\]\] := \[.* .*\]",
        r"  point2\[q_out\.a == x \+ y\] := .*",
    ])
    # print(pattern, problem, sep="\n")
    assert re.match(pattern, repr(problem))


def test_MathematicalProblem_repr_4(s2: System):
    """Test representation of a math problem with targets.
    """
    problem = s2.assembled_problem()
    assert repr(problem) == "\n".join([
        "Equations [5]",
        "  sub.z == 0.12 (target)",
        "  out.a == 3.14 (target)",
        "  abs(out.b) == 2.3 (target)",
        "  out.c[::2] == array([0.1, 0.3]) (target)",
    ])


def test_MathematicalProblem_repr_5(s2: System):
    """Test math problem representation.
    Case: single-point design.
    """
    solver = s2.add_driver(NonLinearSolver('solver'))
    for driver in solver.tree():
        driver.setup_run()
    problem = solver.problem
    assert repr(problem) == "\n".join([
        "Equations [5]",
        "  sub.z == 0.12 := 0.0",
        "  out.a == 3.14 := 0.0",
        "  abs(out.b) == 2.3 := 0.0",
        "  out.c[::2] == array([0.1, 0.3]) := [0. 0.]",
    ])


def test_MathematicalProblem_repr_6(s2: System):
    """Test math problem representation.
    Case: multi-point design with targets.
    """
    solver = s2.add_driver(NonLinearSolver('solver'))
    point1 = solver.add_driver(RunSingleCase('point1'))
    point2 = solver.add_driver(RunSingleCase('point2'))

    point1.set_init({
        'sub.z': 0.12,
        'out.a': 3.14,
        'out.b': -7.3,
        'out.c': np.r_[0.1, 0.2, 0.3, 0.4],
    })
    point2.set_init({
        'sub.z': 0.0,
        'out.a': 0.5,
        'out.b': 0.0,
        'out.c': np.ones(4),
    })

    for driver in solver.tree():
        driver.setup_run()

    # print(solver.problem)
    assert repr(solver.problem) == "\n".join([
        "Equations [10]",
        #  point1 equations
        "  point1[sub.z == 0.12] := 0.0",
        "  point1[out.a == 3.14] := 0.0",
        "  point1[abs(out.b) == 7.3] := 0.0",
        "  point1[out.c[::2] == array([0.1, 0.3])] := [0. 0.]",
        #  point2 equations
        "  point2[sub.z == 0.0] := 0.0",
        "  point2[out.a == 0.5] := 0.0",
        "  point2[abs(out.b) == 0.0] := 0.0",
        "  point2[out.c[::2] == array([1., 1.])] := [0. 0.]",
    ])



import json
import logging
import os, re
from io import StringIO
from pathlib import Path
from tempfile import gettempdir
from unittest import mock

import pytest
from jsonschema import ValidationError
from typing import Callable

from cosapp.core import config
from cosapp.core.config import CoSAppConfiguration

import functools


here = os.path.dirname(os.path.abspath(config.__file__))


def patch_env(env: dict):
    """Decorator reversibly patching `os.environ`, with no side effect."""
    original = os.environ.copy()
    def deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            os.environ.update(env)
            try:
                return func(*args, **kwargs)
            finally:
                os.environ = original
        return wrapper
    return deco


def matcher(pattern: str) -> Callable[[str], bool]:
    """Utility function returning a match test function."""
    return lambda message: re.match(pattern, message)


@pytest.fixture
def fake_schema():
    with open(os.path.join(here, "configuration_schema.json")) as f:
        schema = f.read()

    class FakeSchema:
        def __init__(self, *args, **kwargs):
            self.f = StringIO(schema)

        def __enter__(self):
            return self.f

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.f.close()

    return FakeSchema


def test_CoSAppConfiguration__init__(fake_schema):
    fake_param_str = json.dumps(
        {
            "userid": "id1234",
            "roles": [
                ["scope1", "scope2"],
                ["scope3", "scope4"],
            ],
        }
    )
    fake_io = StringIO(fake_param_str)

    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        config = CoSAppConfiguration()
        assert config.userid == "id1234"
        assert {"scope1", "scope2"} in config.roles
        assert {"scope3", "scope4"} in config.roles


def test_CoSAppConfiguration_userid():
    config = CoSAppConfiguration()
    with pytest.raises(AttributeError):
        config.userid = "r12345"


def test_CoSAppConfiguration_roles():
    config = CoSAppConfiguration()
    with pytest.raises(AttributeError):
        config.roles = [["r12345"]]


def test_CoSAppConfiguration_validate_file(fake_schema):
    # minimal
    fake_io = StringIO(json.dumps({"userid": "a", "roles": []}))

    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        config = CoSAppConfiguration()
        assert config.userid == "a"
        assert len(config.roles) == 0


@pytest.mark.parametrize(
    "configuration",
    [
        ({"userid": "", "roles": [["role1"]]}),  # invalid userid
        ({"roles": [["role1"]]}),  # missing userid
        ({"userid": "abc"}),  # missing roles
        ({"userid": "abc", "roles": [["role1"]], "banana": 42}),  # unexpected keyword
        ({"userid": "abc", "roles": ["role1"]}),  # role is not an array
        ({"userid": "abc", "roles": [["role1", 1]]}),  # role contains none string element
        ({"userid": "abc", "roles": [["role1", "role2"], ["role1", "role2"]]}),  # duplicated role
        ({"userid": "abc", "roles": [["role1", "role2"], ["role3", "role3"]]}),  # duplicated tag in role
        ({"userid": "abc", "roles": [["role1", "role2"], ["role3", ""]]}),  # empty tag string
    ],
)
def test_CoSAppConfiguration_validate_file_error(fake_schema, configuration):
    fake_io = StringIO(json.dumps(configuration))

    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        with pytest.raises(ValidationError):
            CoSAppConfiguration()


@mock.patch.object(CoSAppConfiguration, "update_userid", side_effect=OSError)
def test_CoSAppConfiguration_userid_oserror(fake_schema, caplog):
    """Check that default user ID 'UNKNOWN' is used when
    ID cannot be determined from environment variables.
    This may occur on systems where neither USER nor USERNAME
    environment variables are defined, for instance.

    In this test, method `update_userid` is patched to return `OSError`,
    on top of `open` returning `FileNotFoundError` when trying to access
    user config file.

    Related to: https://gitlab.com/cosapp/cosapp/-/issues/120
    """
    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                raise FileNotFoundError
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        caplog.clear()
        with caplog.at_level(logging.WARNING):
            config = CoSAppConfiguration()
        assert config.userid == "UNKNOWN"
        assert config.roles == frozenset()
        warnings = [
            r.message for r in caplog.records 
            if r.levelno == logging.WARNING
        ]
        assert any(map(
            matcher("Configuration file `.*` cannot be opened"),
            warnings
        ))
        assert any(map(
            matcher("Unable to determine user ID"),
            warnings
        ))


@patch_env({"USER": "", "USERNAME": ""})
def test_CoSAppConfiguration_userid_invalid(fake_schema, caplog):
    """Check that default user ID 'UNKNOWN' is used when
    ID cannot be determined from environment variables.
    This may occur on systems where neither USER nor USERNAME
    environment variables are defined, for instance.

    In this test, `open` is patched to return `FileNotFoundError`
    when trying to access user config file, and environment
    variables USER and USERNAME are set to invalid values.

    Related to: https://gitlab.com/cosapp/cosapp/-/issues/120
    """
    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                raise FileNotFoundError
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        caplog.clear()
        with caplog.at_level(logging.WARNING):
            config = CoSAppConfiguration()
        assert config.userid == "UNKNOWN"
        assert config.roles == frozenset()
        warnings = [
            r.message for r in caplog.records 
            if r.levelno == logging.WARNING
        ]
        assert any(map(
            matcher("Configuration file `.*` cannot be opened"),
            warnings
        ))
        assert any(map(
            matcher("Unable to determine user ID"),
            warnings
        ))


@patch_env({"USER": "", "USERNAME": ""})
def test_CoSAppConfiguration_update_userid_error(fake_schema):

    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                raise FileNotFoundError
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        config = CoSAppConfiguration()
        config._userid = None

        with pytest.raises(OSError, match="Unable to determine user ID"):
            config.update_userid()


def test_CoSAppConfiguration_update_configuration(fake_schema):
    config = {
        "userid": "ab123",
        "roles": [
            ["role1"],
            ["foo", "bar"],
            ["led", "zep"],
        ],
    }
    fake_param_str = json.dumps(config)
    fake_io = StringIO(fake_param_str)

    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        config = CoSAppConfiguration()
        
        # `open` called twice at construction: once to read the config file,
        # the other to read the validation schema, both is read mode.
        assert mock_open.call_count == 2
        call_args = list(map(
            lambda call: call.args,
            mock_open.call_args_list
        ))
        assert str(call_args[0][0]).endswith("cosapp_config.json")
        assert str(call_args[1][0]).endswith("configuration_schema.json")
        assert call_args[0][1] == "r"
        assert call_args[1][1] == "r"

        assert config.userid == "ab123"
        assert config.roles == {
            frozenset({"role1"}),
            frozenset({"foo", "bar"}),
            frozenset({"led", "zep"}),
        }
        assert isinstance(config.roles, frozenset)

        # Reset and update config
        mock_open.reset_mock()
        assert mock_open.call_count == 0
        config.update_configuration()
        assert mock_open.call_count == 1
        call_args = list(map(
            lambda call: call.args,
            mock_open.call_args_list
        ))
        assert str(call_args[0][0]).endswith("cosapp_config.json")
        assert call_args[0][1] == "w"
        assert mock.call.__enter__().write('"userid"') in mock_open.return_value.mock_calls


@patch_env({
    "COSAPP_CONFIG_DIR": str(
        Path(gettempdir()).joinpath("dummy_cosapp")
    ),
})
def test_CoSAppConfiguration_config_var_env(fake_schema):
    folder = Path(os.environ["COSAPP_CONFIG_DIR"])
    init = {"userid": "ab123", "roles": [["role1"]]}
    fake_param_str = json.dumps(init)
    fake_io = StringIO(fake_param_str)

    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fake_schema()
            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        CoSAppConfiguration()

        config_file = folder.joinpath(CoSAppConfiguration.CONFIG_FILE)
        for call in mock_open.calls:
            assert call in [mock.call(config_file, "r"), mock.call(config_file, "w")]


def test_CoSAppConfiguration___init__oserror(caplog):
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        with mock.patch("builtins.open") as mock_open:
            mock_open.side_effect = PermissionError

            CoSAppConfiguration()

    warnings = [r.message for r in caplog.records if r.levelno == logging.WARNING]
    is_match = matcher(
        "Configuration file `.*` cannot be opened; fall back to default"
    )
    assert any(map(is_match, warnings))


def test_CoSAppConfiguration_update_configuration_oserror(caplog):
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        config = CoSAppConfiguration()

        with mock.patch("builtins.open") as mock_open:
            mock_open.side_effect = PermissionError
            config.update_configuration()

    assert "Failed to save configuration locally." in caplog.text



"""
Tool to print the description of a CoSApp object.
"""
from typing import Union

from cosapp.patterns.visitor import Visitor
from cosapp.core.module import Module
from cosapp.ports.enum import PortType
from cosapp.ports.port import BasePort
from cosapp.tools.views.markdown import PortMarkdownFormatter


class DocVisitor(Visitor):
    """Visitor collecting and reformatting Markdown
    representations of systems and ports.
    """
    def __init__(self) -> None:
        self.doc = []

    def visit_port(self, port) -> None:
        """Formatting of port Markdown representation"""
        self.add_doc(port)
        if len(port) > 0:
            formatter = PortMarkdownFormatter(port)
            self.doc.extend(["", "###  Variables", ""])
            self.doc.extend(formatter.var_repr())

    def visit_default(self, obj) -> None:
        """Default formatting of `obj` Markdown representation"""
        self.add_doc(obj)
        self.doc.append(obj._repr_markdown_())

    def visit_system(self, system) -> None:
        """Formatting of system Markdown representation"""
        self.visit_default(system)

    def visit_driver(self, driver) -> None:
        """Formatting of driver Markdown representation"""
        self.visit_default(driver)

    def add_doc(self, obj) -> None:
        """Add header with `obj` class name, and docstring (if any)."""
        obj_type = type(obj)
        doc = self.doc
        doc.extend([f"## Class: {obj_type.__name__}", ""])

        indent = 0
        if obj_type.__doc__:
            doc.extend(["### Documentation", ""])
            for line in obj_type.__doc__.split("\n"):
                if indent == 0:
                    stripped = line.lstrip()
                    if (n_stripped := len(stripped)) > 0:
                        indent = len(line) - n_stripped
                else:
                    stripped = line[indent:]
                doc.append(stripped)


class DocDisplay:
    """Helper class to print nicely information about CoSApp classes."""

    # TODO unit tests

    def __init__(self, obj: Union[type, BasePort, Module], **kwargs):
        """DocDisplay constructor.

        Documentation can only be built for object of type :py:class:`~cosapp.drivers.driver.Driver`,
        :py:class:`~cosapp.ports.port.Port` or :py:class:`~cosapp.systems.system.System`.

        Parameters
        ----------
        - obj: `Port`, `System` or `Driver` class, or instance thereof.
            Object to display.
        - **kwargs:
            Additional keyword arguments forwarded to class constructor,
            if `obj` is a class.
        """
        supported = (BasePort, Module)
        if not (isinstance(obj, supported) or issubclass(obj, supported)):
            raise TypeError("Only Driver, Port and System are supported for display.")

        if isinstance(obj, type):  # Instantiate an object
            if issubclass(obj, BasePort):
                obj = obj("dummy", PortType.OUT, **kwargs)
            else:
                obj = obj("dummy", **kwargs)

        self._obj = obj

    def __repr__(self) -> str:
        """Returns the documentation of the class given to this constructor
        object.

        Returns
        -------
        str
            Documentation string.
        """
        return self._repr_markdown_()

    def _repr_markdown_(self) -> str:
        """Returns the documentation of the class given to this constructor
        object.

        Returns
        -------
        str
            Documentation markdown formatted string.
        """
        visitor = DocVisitor()
        self._obj.accept(visitor)
        return "\n".join(visitor.doc)


def display_doc(obj: Union[type, BasePort, Module], **kwargs) -> DocDisplay:
    """Display information for `Driver`, `System` or `Port` in a Jupyter notebook.

    Parameters
    ----------
    - obj: `Port`, `System` or `Driver` class, or instance thereof.
        Object to display.
    - **kwargs:
        Additional keyword arguments forwarded to class constructor,
        if `obj` is a class.
    """
    return DocDisplay(obj, **kwargs)



"""
Module provides dispatch to different non-linear algorithms.
"""
from __future__ import annotations
import abc
import logging
import inspect
from numbers import Number
from typing import (
    Callable, Optional, Union,
    Sequence, Tuple, Dict, Any,
    TypeVar, Type,
)

import numpy
from scipy.linalg import lu_factor, lu_solve, LinAlgWarning
from scipy import optimize

from cosapp.core.numerics.enum import NonLinearMethods
from cosapp.core.numerics.basics import SolverResults
from cosapp.utils.logging import LogLevel

logger = logging.getLogger(__name__)


ConcreteSolver = TypeVar("ConcreteSolver", bound="BaseNumericSolver")
RootFunction = Callable[[Sequence[float], Any], numpy.ndarray]


class BaseNumericalSolver(abc.ABC):

    def __init__(self, options={}) -> None:
        self.__options = dict(options)

    @property
    def options(self) -> Dict[str, Any]:
        return self.__options

    @classmethod
    def from_options(cls: Type[ConcreteSolver], options: Dict[str, Any]) -> ConcreteSolver:
        solver = cls()
        solver.transfer_options(options)
        return solver

    def transfer_options(self, options: Dict[str, Any]) -> None:
        settings = self.options
        for key in settings.keys():
            try:
                settings[key] = options.pop(key)
            except KeyError:
                continue

    @abc.abstractmethod
    def solve(
        self,
        fun: RootFunction,
        x0: Sequence[float],
        args: Tuple[Any] = tuple(),
        *other_args, **kwargs,
    ) -> Union[SolverResults, optimize.OptimizeResult]:
        pass


class NumpySolver(BaseNumericalSolver):
    """Encapsulation of `scipy.optimize.root`
    """
    def __init__(self, method: str, options={}) -> None:
        """
        Parameters
        ----------
        method: str
            Method forwarded to `scipy.optimize.root`
        options: Dict[str, Any], optional
            A dictionary of solver options. E.g. `tol` or `max_iter`
        """
        self._method = method
        super().__init__(options)
        try:
            self.options.pop('verbose')
        except KeyError:
            pass

    def solve(self,
        fun: RootFunction,
        x0: Sequence[float],
        args: Tuple[Union[float, str]] = (),
        jac=None,
        callback=None,
    ) -> optimize.OptimizeResult:
        """Cancel residues produced by `fun` starting with `x0` as initial guess.

        Parameters
        ----------
        fun : Callable[[Sequence[float], Any], numpy.ndarray[float]]
            Callable residues function
        x0 : Sequence[float]
            Initial solution guess
        args : Tuple[Any, ...]
            Additional arguments passed to `fun`
        jac : bool or callable, optional
            If `jac` is a Boolean and is `True`, `fun` is assumed to return
            the value of Jacobian along with the objective function.
            If `False`, the Jacobian will be estimated numerically.
            `jac` can also be a callable returning the Jacobian of `fun`.
            In this case, it must accept the same arguments as `fun`.
            (default: `None`)
        callback : Callable, optional
            Optional callback function. It is called on every iteration as `callback(x, r)`,
            where x is the current solution and r the corresponding residual.

        Returns
        -------
        scipy.optimize.OptimizeResult
            The solution represented as a OptimizeResult object. Important attributes are: x the solution array,
            success a Boolean flag indicating if the algorithm exited successfully and message which describes the
            cause of the termination. See :py:class:`cosapp.core.numerics.basics.SolverResults` for a description
            of other attributes.
        """
        x0 = numpy.atleast_1d(x0)

        results = optimize.root(
            fun, x0, args=args,
            method=self._method,
            jac=jac,
            tol=self.options.get('tol', None),
            callback=callback,
            options=self.options,
        )
        results.jac_lup = (None, None)
        if 'jac' in results:
            try:
                results.jac_lup = lu_factor(results.jac, check_finite=False)
            except (ValueError, LinAlgWarning) as err:  # Silent LU decomposition failure
                logger.debug(f"Silent error: {err}")

        return results


def get_kwargs():
    """Returns kwargs as a dictionary, using `inspect.currentframe`
    """
    frame = inspect.currentframe().f_back
    keys, _, _, values = inspect.getargvalues(frame)
    kwargs = {}
    for key in keys:
        if key != 'self':
            kwargs[key] = values[key]
    return kwargs


class CustomSolver(BaseNumericalSolver):
    """Custom Newton-Raphson solver.
    """
    def __init__(
        self,
        tol='auto',
        factor=1.0,
        max_iter=100,
        verbose=False,
        eps=2**(-23),
        jac_update_tol=0.01,
        history=False,
        partial_jac=True,
        partial_jac_tries=5,
        tol_update_period=4,
        tol_to_noise_ratio=16,
    ):
        options = get_kwargs()
        super().__init__(options)

    def solve(
        self,
        fresidues: RootFunction,
        x0: Sequence[float],
        args: Tuple[Union[float, str]] = (),
        jac=None,
        recorder=None,
        **options,
    ) -> SolverResults:
        """Customized Newton-Raphson algorithm to solve `fresidues` starting with `x0`.

        Parameters
        ----------
        fresidues : Callable[[Sequence[float], Any], numpy.ndarray[float]]
            Callable residues function
        x0 : Sequence[float]
            Initial solution guess
        args : Tuple[Any, ...]
            Additional arguments for `fun`
        options : dict, optional
            A dictionary of problem-dependent solver options.

        Options
        -------
        jac_lup : (ndarray[float], ndarray[int]), optional
            LU decomposition of Jacobian given as tuple (LU, perm) to reuse as initial direction
        jac : ndarray, optional
            Jacobian to reuse for partial update
        compute_jacobian : bool
            Force to update the Jacobian matrix even if the provided one is useful
        lower_bound : numpy.ndarray
            Min values for parameters iterated by solver.
        upper_bound : numpy.ndarray
            Max values for parameters iterated by solver.
        abs_step : numpy.ndarray
            Max absolute step for parameters iterated by solver.
        rel_step : numpy.ndarray
            Max relative step for parameters iterated by solver.

        Returns
        -------
        SolverResults
            The solution represented as a `SolverResults` object.
            See :py:class:`cosapp.core.numerics.basics.SolverResults` for details.
        """
        solver_options = self.options
        verbose = solver_options.get('verbose', False)
        log_level = LogLevel.INFO if verbose else LogLevel.DEBUG
        jac_update_tol = solver_options.get('jac_update_tol', 0.01)
        jac_rel_perturbation = solver_options['eps']
        factor_ref = factor = solver_options['factor']
        history = solver_options.get('history', False)

        if jac is not None:
            jac = numpy.asarray(jac)
        jac_lu, piv = options.get('jac_lup', (None, None))
        calc_jac = options.get('compute_jacobian', True) or jac is None or jac.shape[1] != x0.shape[0]

        def calc_jacobian(
            x: numpy.ndarray,
            residues: numpy.ndarray,
            it_jac: int,
            it_p_jac: int,
            jac: numpy.ndarray=None,
            r_indices_to_update=set(),
        ) -> Tuple[numpy.ndarray, int, int]:
            """Estimate function gradient with respect to x"""
            n = x.size
            if jac is None or jac.shape != (n, n):
                new_jac = numpy.zeros((n, n), dtype=float)
                x_indices_to_update = list(range(n))
            else:
                new_jac = jac.copy()
                ncol = new_jac.shape[1]
                x_indices_to_update = set()
                for i in r_indices_to_update:
                    x_indices_to_update.update(j for j in range(ncol) if new_jac[i, j] != 0)

            x_copy = x.copy()

            for j in x_indices_to_update:
                delta = jac_rel_perturbation
                if abs(x_copy[j]) >= abs(jac_rel_perturbation):
                    delta = x_copy[j] * jac_rel_perturbation
                x_copy[j] += delta
                logger.debug(f"Perturb unknown {j}")
                perturbation = fresidues(x_copy, *args)
                new_jac[:, j] = (perturbation - residues) * (1 / delta)
                x_copy[j] = x[j]

            n_partial = len(x_indices_to_update)
            if n_partial == n:
                logger.log(log_level, f'Jacobian matrix: full update')
                it_jac += 1
            elif x_indices_to_update:
                it_p_jac += 1
                logger.log(log_level, 
                    f'Jacobian matrix: {n_partial} over {n} derivative(s) updated')

            return new_jac, it_jac, it_p_jac

        logger.debug("NR - Reference call")
        x = numpy.array(x0, dtype=numpy.float64).flatten()
        r = fresidues(x, *args)
        r_norm = numpy.linalg.norm(r, numpy.inf)
        logger.log(log_level, f"Initial residue: {r_norm}")
        logger.log(LogLevel.FULL_DEBUG, f"Initial relaxation factor {factor}.")
        results = SolverResults()

        def check_numerical_features(name, default_value):
            if name not in options or options[name] is None or numpy.shape(options[name]) != x.shape:
                options[name] = numpy.full_like(x, default_value)

        check_numerical_features('lower_bound', -numpy.inf)
        check_numerical_features('upper_bound', numpy.inf)
        check_numerical_features('abs_step', numpy.inf)
        check_numerical_features('rel_step', numpy.inf)

        it_solver = 0
        if recorder is not None:
            record_state = lambda iter: recorder.record_state(str(iter), 'ok',)
        else:
            record_state = lambda iter: None

        if not calc_jac:
            logger.debug('Reuse of previous Jacobian matrix')

        tol = solver_options['tol']
        if tol is None:
            tol = 'auto'
        elif isinstance(tol, str):
            if tol != 'auto':
                raise ValueError(f"Tolerance must be a float, None, or 'auto'.")
        elif not isinstance(tol, Number):
            raise TypeError(f"Tolerance must be a number, None, or 'auto'; got {tol!r}")
        
        auto_tol = (tol == 'auto')
        if auto_tol:
            tol = 0.0
            iter_period = solver_options['tol_update_period']
            tol_to_noise = solver_options['tol_to_noise_ratio']
            must_update_tol = lambda it: it % iter_period == 0
        elif tol < 0:
            raise ValueError(f"Tolerance must be non-negative (got {tol})")
        else:
            tol_to_noise = None
            must_update_tol = lambda it: False

        results.trace = trace = list()
        if history:
            record = {
                "x": x.copy(),
                "residues": r.copy(),
                "tol": tol,
            }
            trace.append(record)

        max_iter = solver_options['max_iter']
        p_jac = solver_options['partial_jac']
        p_jac_tries = solver_options['partial_jac_tries']
        it_jac, it_p_jac = 0, 0  # number of full and partial evalutions of the Jacobian
        n_broyden_update = 0
        dr = numpy.zeros_like(r)
        dx = numpy.full_like(r, numpy.nan)

        rtol_x = 1e-14
        epsilon = numpy.finfo(numpy.float64).eps
        logger.log(
            LogLevel.FULL_DEBUG,
            "\t".join([
                f"iter #{it_solver}",
                f"tol = {tol:.2e}",
                f"|R| = {r_norm}",
                f"x = {x}",
            ])
        )

        try:
            res_index_to_update = set()
            consecutive_p_jac = 0
            prev_tol = -1.0

            while r_norm > tol and it_solver < max_iter:
                logger.log(log_level, f'Iteration {it_solver}')

                if calc_jac:
                    if consecutive_p_jac > p_jac_tries or not p_jac:
                        res_index_to_update = range(x.size)
                        consecutive_p_jac = 0
                    else:
                        consecutive_p_jac += 1

                    jac, it_jac, it_p_jac = calc_jacobian(
                        x, r, it_jac, it_p_jac,
                        jac, res_index_to_update,
                    )
                    jac_lu, piv = self.lu_factor(jac)  # may raise an exception

                elif it_solver > 0:
                    # Good Broyden update - source: https://nickcdryan.com/2017/09/16/broydens-method-in-python/
                    logger.log(log_level, f'Broyden update')
                    n_broyden_update += 1
                    corr = numpy.outer((dr - jac.dot(dx)), dx) / dx.dot(dx)
                    jac += corr
                    jac_lu, piv = self.lu_factor(jac)

                if must_update_tol(it_solver):
                    norm_J = numpy.linalg.norm(jac, numpy.inf)
                    norm_x = numpy.linalg.norm(x, numpy.inf)
                    noise = epsilon * norm_J * norm_x
                    tol = tol_to_noise * noise
                    logger.log(
                        LogLevel.FULL_DEBUG,
                        f"iter #{it_solver}; noise level = {noise}; tol = {tol}; |J| = {norm_J}; |x| = {norm_x}"
                    )
                    if tol == prev_tol:
                        logger.log(log_level, f"Numerical saturation detected at iteration {it_solver}; x = {x}")
                        break
                    prev_tol = tol

                dx = -lu_solve((jac_lu, piv), r)
                it_solver += 1

                # Compute relaxation factors from max absolute and relative steps
                with numpy.errstate(invalid='ignore', divide='ignore'):
                    abs_x = numpy.abs(x)
                    abs_dx = numpy.abs(dx)
                    dx_max_abs = options['abs_step']
                    dx_max_rel = numpy.where(abs_x > 0, abs_x * options['rel_step'], numpy.inf)
                    factor_abs = numpy.where(abs_dx > dx_max_abs, dx_max_abs / abs_dx, 1).min()
                    factor_rel = numpy.where(abs_dx > dx_max_rel, dx_max_rel / abs_dx, 1).min()

                factor = min(factor_rel, factor_abs, factor, 1)

                if factor < factor_ref:
                    logger.debug(f"\trelaxation factor lowered to {factor}")

                r_norm_prev = r_norm
                x_prev = x.copy()
                dx *= factor
                x += dx

                # Force solution within admissible bounds
                # Note:
                #   This is clearly wrong, and does NOT belong in a nonlinear solver.
                #   The correct way to deal with such constraints is to use an optimizer.
                #   This part was only kept for non-regression purposes, but should be
                #   removed in a future revision.
                x = numpy.maximum(x, options['lower_bound'])
                x = numpy.minimum(x, options['upper_bound'])

                new_r = fresidues(x, *args)
                r_norm = numpy.linalg.norm(new_r, numpy.inf)
                # logger.log(log_level, f'Residue: {r_norm:.5g}')
                logger.log(
                    LogLevel.DEBUG,
                    "\t".join([
                        f"iter #{it_solver}",
                        f"tol = {tol:.2e}",
                        f"|R| = {r_norm}",
                        f"x = {x}",
                    ])
                )

                if numpy.allclose(x, x_prev, rtol=rtol_x, atol=0):
                    logger.log(log_level, f"Fixed point detected: x = {x}, dx = {x - x_prev}")
                    break

                # Estimate non-linearity by comparing extrapolated and actual residues
                abs_r = numpy.abs(r)
                extrapolated_r = r + jac.dot(dx)
                delta_vs_linear = numpy.abs(extrapolated_r - new_r)
                res_index_to_update = set(
                    numpy.argwhere(delta_vs_linear > jac_update_tol * abs_r).ravel()
                )
                res_index_to_update -= set(numpy.argwhere(abs_r < tol).ravel())

                r, dr = new_r, new_r - r

                if history:
                    record = {
                        "x": x.copy(),
                        "residues": r.copy(),
                        "tol": tol,
                    }
                    if calc_jac:
                        record["jac"] = jac.copy()
                    trace.append(record)
                
                record_state(it_solver)

                if res_index_to_update:
                    factor = factor_ref
                    calc_jac = True
                elif r_norm != 0:
                    growth_rate = 1 + 0.1 * (1 - factor) * r_norm_prev / r_norm
                    factor = min(growth_rate * factor, 1)
                    calc_jac = False
                    logger.log(LogLevel.FULL_DEBUG, f"New relaxation factor {factor}.")

        except (ValueError, LinAlgWarning):
            is_nil = (jac == 0)
            zero_rows = numpy.argwhere(is_nil.all(axis=1)).flatten()
            zero_cols = numpy.argwhere(is_nil.all(axis=0)).flatten()
            results.success = False
            results.message = 'Singular {}x{} Jacobian matrix'.format(*jac.shape)
            results.jac_errors = {'unknowns': zero_cols, 'residues': zero_rows}
        
        else:
            results.x = x
            results.jac_lup = (jac_lu, piv)
            results.jac_calls = it_jac + it_p_jac
            results.fres_calls = it_solver
            results.success = (r_norm <= tol)
            
            status = f'Converged' if results.success else 'Not converged'
            message = (
                f"{status} ({r_norm:.4e}) in {it_solver} iterations,"
                f" {it_jac} complete, {it_p_jac} partial Jacobian and"
                f" {n_broyden_update} Broyden evaluation(s)"
                f" (tol = {tol:.1e})"
            )
            results.message = f"   -> {message}"

        finally:
            results.jac = jac
            results.tol = tol
            return results

    @staticmethod
    def lu_factor(matrix: numpy.ndarray):
        lu, piv = lu_factor(matrix, check_finite=True)
        min_diag = numpy.abs(lu.diagonal()).min()
        if min_diag < 1e-14:
            raise LinAlgWarning(
                f"Quasi-singular Jacobian matrix; min diag element of U matrix is {min_diag}"
            )
        return lu, piv


def root(
    fun: RootFunction,
    x0: Sequence[float],
    args: Tuple[Any] = tuple(),
    method: NonLinearMethods = NonLinearMethods.POWELL,
    options: Dict[str, Any] = {},
) -> Union[SolverResults, optimize.OptimizeResult]:

    if method == NonLinearMethods.NR:
        solver = CustomSolver.from_options(options)
        return solver.solve(fun, x0, args, **options)
    
    else:
        solver = NumpySolver(method.value, options)
        return solver.solve(fun, x0, args)



"""This module defines the basic class encapsulating mode variable attributes."""
import copy
from typing import Any, Optional

from cosapp.ports.enum import Scope
from cosapp.ports.variable import BaseVariable, Types
from cosapp.utils.helpers import check_arg

import logging
logger = logging.getLogger(__name__)


class ModeVariable(BaseVariable):
    """Container for mode variables.

    Parameters
    ----------
    name : str
        Variable name
    port : ModeVarPort, optional
        Port to which the variable belongs
    value : Any
        Variable value
    unit : str, optional
        Variable unit; default empty string (i.e. dimensionless)
    dtype : type or iterable of type, optional
        Variable type; default None (i.e. type of initial value)
    desc : str, optional
        Variable description; default ''
    init : Any, optional
        Value imposed at the beginning of time simulations, if
        variable is an output (unused otherwise).
        If unspecified (default), the variable remains untouched.
    scope : Scope {PRIVATE, PROTECTED, PUBLIC}, optional
        Variable visibility; default PRIVATE
    """

    __slots__ = (
        "_init",
    )

    def __init__(
        self,
        name: str,
        port: "cosapp.ports.ModeVarPort",
        value: Optional[Any] = None,
        unit: str = "",
        dtype: Types = None,
        desc: str = "",
        init: Optional[Any] = None,
        scope: Scope = Scope.PRIVATE,
    ):
        from cosapp.ports import ModeVarPort
        from cosapp.core.eval_str import EvalString
        check_arg(port, "port", ModeVarPort)
        init = EvalString(init, port.owner)
        init_value = init.eval()
        if value is None:
            value = init_value
        elif init_value is not None:
            if not isinstance(init_value, type(value)):
                raise TypeError(
                    f"Initial value {init!r} appears to be inconsistent with arg value={value!r}."
                )

        super().__init__(name, port, value, unit, dtype, desc, scope)

        self._init = init  # type: EvalString

    def _repr_markdown_(self) -> str:
        """Returns the representation of this variable in Markdown format.

        Returns
        -------
        str
            Markdown formatted representation
        """
        msg = {"name":f"**{self.name}**" , "unit": f" {self.unit}" if self.unit else ""}
        value = self.value
        try:
            msg["value"] = f"{value:.5g}"
        except:
            msg["value"] = value

        lock_icon = "&#128274;"
        if self.description:
            msg["description"] = f" | {self.description}"
        else:
            msg["description"] = " |"

        scope_format = {
            Scope.PRIVATE: f" {lock_icon*2} ",
            Scope.PROTECTED: f" {lock_icon} ",
            Scope.PUBLIC: "",
        }
        msg["scope"] = scope_format[self.scope]

        return (
            "{name}{scope}: {value!s}{unit}"
            "{description}".format(**msg)
        )  

    @property
    def init_expr(self) -> "cosapp.core.eval_str.EvalString":
        """EvalString : expression of initial value"""
        return self._init

    def init_value(self) -> Any:
        """Evaluate and return initial value"""
        return self._init.eval()

    def initialize(self) -> None:
        """Set mode variable to its prescribed initial value."""
        value = self.init_value()
        if value is not None:
            try:
                setattr(self._port, self._name, value)
            except AttributeError:
                pass

    def copy(self, port: "BasePort", name: Optional[str] = None) -> "ModeVariable":
        if name is None:
            name = self.name
        return ModeVariable(
            name,
            value = copy.copy(self.value),
            port = port,
            unit = self._unit,
            dtype = copy.copy(self._dtype),
            desc = self._desc,
            scope = self._scope,
        )



import pytest
import numpy as np
from cosapp import drivers

from cosapp.systems import System
from cosapp.drivers import RungeKutta
from cosapp.recorders import DataFrameRecorder


class Pendulum(System):
    def setup(self):
        self.add_inward('L', 1.00, desc='Length of the rod')
        self.add_inward('g', 9.81, desc='Gravitational acceleration')
        self.add_outward('acc', 1., desc="Angular acceleration of the mass")
        
        self.add_transient('omega', der='acc', desc="Angular velocity")
        self.add_transient('theta', der='omega', desc="Positional angle")

    def compute(self):
        self.acc = -(self.g / self.L) * np.sin(self.theta)


class NewtonPendulum(System):
    def setup(self):
        self.add_child(Pendulum('p1'), pulling=['L', 'g'])
        self.add_child(Pendulum('p2'), pulling=['L', 'g'])
        
        contact = self.add_event('contact', trigger="p1.theta == p2.theta")
        self.add_event('collision', trigger=contact.filter("p1.omega > p2.omega"))

    def transition(self):
        if self.collision.present:
            self.p1.omega, self.p2.omega = self.p2.omega, self.p1.omega


def test_NewtonPendulum():
    pend = NewtonPendulum("pend")

    driver = pend.add_driver(
        RungeKutta(order=3, dt=0.01, time_interval=[0, 2])
    )
    driver.set_scenario(
        init = {
            "p1.theta" : np.radians(-60),
            "p1.omega" : 2,
            "p2.theta" : np.radians(22),
            "p2.omega" : 0.5,
        },
        values = {
            "L" : 0.5,
            "g" : 9.81,
        },
    )
    driver.add_recorder(
        DataFrameRecorder(includes=['p?.theta', 'p?.omega']),
        period = 0.05,
    )

    # Run simulation
    pend.run_drivers()

    data = driver.recorder.export_data()
    theta1 = np.asarray(data['p1.theta'])
    theta2 = np.asarray(data['p2.theta'])
    # Check that theta1 <= theta2 within numerical tolerance:
    assert all(theta1 - theta2 < 1e-15)
    # Check recorded events
    recorded_events = driver.recorded_events
    assert len(recorded_events) == 3
    assert [record.time for record in recorded_events] == pytest.approx(
        [0.31293088163661, 1.0724999600884, 1.831490584856]
    )
    for record in recorded_events:
        assert len(record.events) == 2
        assert record.events[0] is pend.contact  # first in record is primary event
        assert set(record.events) == {pend.contact, pend.collision}


def test_NewtonPendulum_limit():
    """Limit case where the two masses move jointly (constant contact)."""
    pend = NewtonPendulum("pend")

    driver = pend.add_driver(
        RungeKutta(order=3, dt=0.01, time_interval=[0, 1])
    )
    driver.set_scenario(
        init = {
            "p1.theta" : np.radians(10),
            "p1.omega" : -2,
            "p2.theta" : np.radians(10),
            "p2.omega" : -2,
        },
        values = {
            "L" : 0.5,
            "g" : 9.81,
        },
    )
    driver.add_recorder(
        DataFrameRecorder(includes=['p?.theta', 'p?.omega']),
        period = 0.05,
    )

    # Run simulation
    pend.run_drivers()

    # Check that no event was recorded
    assert len(driver.event_data) == 0
    assert len(driver.recorded_events) == 0

    data = driver.recorder.export_data()
    # print(data.drop(['Section', 'Status', 'Error code'], axis=1))
    theta1 = np.asarray(data['p1.theta'])
    theta2 = np.asarray(data['p2.theta'])
    # Check that theta1 == theta2:
    assert theta1 == pytest.approx(theta2, abs=1e-15)


def test_NewtonPendulum_stop():
    """Same as base test, with a stop criterion based on a filtered event."""
    pend = NewtonPendulum("pend")

    driver = pend.add_driver(
        RungeKutta(order=3, dt=0.01, time_interval=[0, 2])
    )
    driver.set_scenario(
        init = {
            "p1.theta" : np.radians(-60),
            "p1.omega" : 2,
            "p2.theta" : np.radians(22),
            "p2.omega" : 0.5,
        },
        values = {
            "L" : 0.5,
            "g" : 9.81,
        },
        stop = pend.contact.filter('p1.theta < 0'),
    )
    driver.add_recorder(
        DataFrameRecorder(includes=['p?.theta', 'p?.omega']),
        period = 0.05,
    )

    # Run simulation
    pend.run_drivers()

    data = driver.recorder.export_data()
    theta1 = np.asarray(data['p1.theta'])
    theta2 = np.asarray(data['p2.theta'])
    # Check that theta1 <= theta2 within numerical tolerance:
    assert all(theta1 - theta2 < 1e-15)
    assert pend.p1.theta == pytest.approx(pend.p2.theta, abs=1e-15)
    assert pend.p1.theta == pytest.approx(-0.1186766797)
    # Check recorded events
    recorded_events = driver.recorded_events
    assert len(recorded_events) == 2

    # First event record
    record = recorded_events[0]
    assert record.time == pytest.approx(0.31293088163661)
    assert len(record.events) == 2
    assert record.events[0] is pend.contact  # first in record is primary event
    assert set(record.events) == {pend.contact, pend.collision}

    # Second event record
    record = recorded_events[1]
    assert record.time == pytest.approx(1.0724999600884)
    assert len(record.events) == 3
    assert record.events[0] is pend.contact  # first in record is primary event
    assert set(record.events) == {pend.contact, pend.collision, driver.scenario.stop}



"""
Surrogate model based on the N-Dimensional Interpolation library by Stephen Marone.

https://github.com/SMarone/NDInterp
"""

import numpy
from typing import Dict, Type
from .base import SurrogateModel
from .nn_interpolators.nn_base import NNBase
from .nn_interpolators.linear_interpolator import LinearInterpolator
from .nn_interpolators.weighted_interpolator import WeightedInterpolator
from .nn_interpolators.rbf_interpolator import RBFInterpolator


class NearestNeighbor(SurrogateModel):
    """
    Surrogate model that approximates values using a nearest neighbor approximation.

    Attributes
    ----------
    interpolant : object
        Interpolator object
    options : dict
        Input keyword arguments for the interpolator.
    interpolant_type : NNBase
        Type of interpolator from NearestNeighbor.interpolators()
    """

    def __init__(self, interpolant_type="rbf", **options):
        """
        Initialize all attributes.

        Parameters
        ----------
        interpolant_type : str
            must be one of 'linear', 'weighted', or 'rbf'.
        **options :
            Options as keyword arguments
        """
        interpolators = self.interpolators()
        try:
            self.interpolant_type = interpolators[interpolant_type]
        except KeyError:
            raise ValueError(
                f"interpolant_type '{interpolant_type}' not supported"
                f"; must be one of {list(interpolators)}."
            )
        self.options = options
        self.interpolant = None

    def train(self, x, y):
        """
        Train the surrogate model with the given set of inputs and outputs.

        Parameters
        ----------
        x : array-like
            Training input locations
        y : array-like
            Model responses at given inputs.
        """
        self.interpolant = self.interpolant_type(
            numpy.asarray(x),
            numpy.asarray(y),
            **self.options,
        )

    def predict(self, x, **kwargs):
        """
        Calculate a predicted value of the response based on the current trained model.

        Parameters
        ----------
        x : array-like
            Point(s) at which the surrogate is evaluated.
        **kwargs : dict
            Additional keyword arguments passed to the interpolant.

        Returns
        -------
        float
            Predicted value.
        """
        return self.interpolant(numpy.asarray(x), **kwargs)

    def linearize(self, x, **kwargs):
        """
        Calculate the jacobian of the interpolant at the requested point.

        Parameters
        ----------
        x : array-like
            Point at which the surrogate Jacobian is evaluated.
        **kwargs : dict
            Additional keyword arguments passed to the interpolant.

        Returns
        -------
        ndarray
            Jacobian of surrogate output wrt inputs.
        """
        jac = self.interpolant.gradient(numpy.asarray(x), **kwargs)
        if jac.shape[0] == 1 and len(jac.shape) > 2:
            return jac[0, ...]
        return jac

    @staticmethod
    def interpolators() -> Dict[str, Type[SurrogateModel]]:
        return {
            "linear": LinearInterpolator,
            "weighted": WeightedInterpolator,
            "rbf": RBFInterpolator,
        }


class LinearNearestNeighbor(NearestNeighbor):  # pragma: no cover
    """
    Surrogate model that approximates values using a linear interpolator.

    Attributes
    ----------
    interpolant : object
        Interpolator object
    options : dict
        Input keyword arguments for the interpolator.
    interpolant_type : Type[NNBase]
        Type of interpolator; here `LinearInterpolator`

    """

    def __init__(self, **options):
        """
        Initialize all attributes.

        Parameters
        ----------
        **options :
            Options, as keyword arguments
        """
        super().__init__("linear", **options)


class WeightedNearestNeighbor(NearestNeighbor):  # pragma: no cover
    """
    Surrogate model that approximates values using a weighted interpolator.

    Attributes
    ----------
    interpolant : object
        Interpolator object
    options : dict
        Input keyword arguments for the interpolator.
    interpolant_type : Type[NNBase]
        Type of interpolator; here `WeightedInterpolator`

    """

    def __init__(self, **options):
        """
        Initialize all attributes.

        Parameters
        ----------
        **options :
            Options, as keyword arguments
        """
        super().__init__("weighted", **options)


class RBFNearestNeighbor(NearestNeighbor):  # pragma: no cover
    """
    Surrogate model that approximates values using a RBF interpolator.

    Attributes
    ----------
    interpolant : object
        Interpolator object
    options : dict
        Input keyword arguments for the interpolator.
    interpolant_type : Type[NNBase]
        Type of interpolator; here `RBFInterpolator`

    """

    def __init__(self, **options):
        """
        Initialize all attributes.

        Parameters
        ----------
        **options :
            Options, as keyword arguments
        """
        super().__init__("rbf", **options)



"""
Copyright (c) 2016-2018, openmdao.org

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this software except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This package comes from OpenMDAO 2.2.0. It was modified for CoSApp integration.
"""

from.base import (
    SurrogateModel,
    MultiFiSurrogateModel,
)
from .kriging import (
    KrigingSurrogate,
    FloatKrigingSurrogate,
)
from .multifi_cokriging import (
    MultiFiCoKrigingSurrogate,
    FloatMultiFiCoKrigingSurrogate,
)
from .nearest_neighbor import (
    NearestNeighbor,
    LinearNearestNeighbor,
    WeightedNearestNeighbor,
    RBFNearestNeighbor,
)
from .response_surface import ResponseSurface

__all__ = [
    "SurrogateModel",
    "MultiFiSurrogateModel",
    "KrigingSurrogate",
    "FloatKrigingSurrogate",
    "MultiFiCoKrigingSurrogate",
    "FloatMultiFiCoKrigingSurrogate",
    "NearestNeighbor",
    "LinearNearestNeighbor",
    "WeightedNearestNeighbor",
    "RBFNearestNeighbor",
    "ResponseSurface",
]


"""
Ports package of CoSApp.
"""
from .enum import PortType, Scope, Validity
from .exceptions import ScopeError
from .connectors import Connector, ConnectorError
from .port import Port, ModeVarPort
from .units import UnitError, add_offset_unit, add_unit, convert_units

__all__ = [
    "ModeVarPort",
    "Port",
    "PortType",
    "Connector",
    "ConnectorError",
    "Scope",
    "ScopeError",
    "Validity",
    "UnitError",
    "add_unit",
    "add_offset_unit",
    "convert_units",
]



import numpy
import collections
import pickle
import pandas
import logging
import warnings
from numbers import Number
from collections import OrderedDict
from typing import (
    NoReturn, Any, Dict, List,
    Iterable, NamedTuple, Union,
    Optional, Type,
)

from cosapp.utils.surrogate_models.base import SurrogateModel
from cosapp.ports.enum import PortType
from cosapp.utils.logging import LogLevel
from cosapp.utils.helpers import check_arg
from cosapp.utils.find_variables import find_variables, make_wishlist, natural_varname


logger = logging.getLogger(__name__)


class SurrogateModelProxy(SurrogateModel):
    """Surrogate model proxy used in `SystemSurrogate`.
    Adds an internal property `trained`, and check that model
    has been trained before each prediction.
    """
    def __init__(self, model):
        if not isinstance(model, SurrogateModel):
            raise TypeError(
                "`SurrogateModelProxy` can only wrap `SurrogateModel` instances."
            )
        self.__wrappee = model
        self.__trained = False

    @property
    def trained(self) -> bool:
        return self.__trained

    def train(self, x, y) -> None:
        """Trains surrogate model with inputs `x` and outputs `y`.

        Parameters
        ----------
        x : array-like
            Training input locations
        y : array-like
            Model responses at given inputs.
        """
        model = self.__wrappee
        logger.debug(f"Training model {model} with \n\tX = {x}\n\tY = {y}")
        model.train(x, y)
        self.__trained = True
        logger.debug(f"Model {model} trained")

    def predict(self, x) -> numpy.ndarray:
        """Calculates a predicted value of the response based on the current trained model.

        Parameters
        ----------
        x : array-like
            Point(s) at which the surrogate is evaluated.
        """
        model = self.__wrappee
        if not self.__trained:
            raise RuntimeError(
                f"{type(model).__name__} has not been trained, so no prediction can be made."
            )
        return model.predict(x)

    def get_type(self) -> Type[SurrogateModel]:
        """Returns wrapped model type"""
        return type(self.__wrappee)


class SystemSurrogateState(NamedTuple):
    """Named tuple containing internal state data of a SystemSurrogate object"""
    doe_in: pandas.DataFrame
    doe_out: Union[pandas.DataFrame, OrderedDict]
    model: SurrogateModelProxy
    doe_out_sizes: OrderedDict


class SystemSurrogate:
    """
    Class representing a surrogate model trained from the outputs of an owner system.
    Once the surrogate is created, it supersedes (whenever activated) the behaviour of
    `System.compute()`. 
    """
    __slots__ = ('__owner', '__state', '__need_doe')

    def __init__(self,
        owner: "cosapp.systems.System",
        data_in: Union[pandas.DataFrame, Dict[str, Any]],
        model: Type[SurrogateModel],
        data_out: Optional[Union[pandas.DataFrame, Dict[str, Any]]] = None,
        postsynch: Union[str, List[str]] = '*',
        *args, **kwargs
    ):
        # CHECK ARGS
        from cosapp.systems import System
        check_arg(owner, 'owner', System)
        check_arg(data_in, 'data_in', (pandas.DataFrame, dict))
        check_arg(data_out, 'data_out', (pandas.DataFrame, dict, type(None)))
        if model is not None:
            try:
                check_arg(model, 'model', type, lambda m: issubclass(m, SurrogateModel))
            except ValueError as error:
                error.args = (
                    f"`model` must be a concrete implementation of `SurrogateModel`"
                    f"; got `{model.__name__}`",
                )
                raise
        postsynch = make_wishlist(postsynch, 'postsynch')
        
        # BUILD INTERN DATA
        self.__owner: System = owner
        self.__need_doe = (data_out is None)
        
        doe_in = pandas.DataFrame.from_dict(data_in) if isinstance(data_in, dict) else data_in
        doe_out = self.__init_doe_out(data_out, postsynch)
        self.filter_headers(doe_in)
        model_obj = None if model is None else SurrogateModelProxy(model(*args, **kwargs))
        self.__state = SystemSurrogateState(doe_in, doe_out, model_obj, OrderedDict())

        # TRAINING
        empty = owner is None or model_obj is None or len(data_in) == 0
        if not empty:
            logger.debug(
                f"Initialize {model.__name__} surrogate for System {owner.name!r}"
                f", with {len(doe_in)} samples of {list(doe_in.columns)}"
            )
            self.__check_unknowns_and_transients()
            self.__prepare_and_train()

        logger.debug(f"System surrogate initialized")

    @property
    def owner(self)-> "cosapp.systems.System":
        return self.__owner

    @property
    def state(self)-> SystemSurrogateState:
        """SystemSurrogateState: inner state of system surrogate"""
        return self.__state

    @property
    def model_type(self)-> Type[SurrogateModel]:
        return self.__state.model.get_type()

    @property
    def trained(self) -> bool:
        return self.__state.model.trained

    @property
    def synched_outputs(self) -> List[str]:
        """List[str]: list of synchronized output variable names"""
        return list(self.__state.doe_out_sizes)

    @staticmethod
    def filter_headers(doe: pandas.DataFrame) -> None:
        """Apply `natural_varname` to dataframe column names"""
        mapping = dict(
            (name, natural_varname(name))
            for name in doe.columns
        )
        doe.rename(mapping, axis=1, inplace=True)

    def __init_doe_out(self, data_out, postsynch) -> Dict[str, List]:
        doe_out = OrderedDict()
        if isinstance(data_out, pandas.DataFrame):
            doe_out = data_out.to_dict(into=OrderedDict)
        elif data_out is not None:
            doe_out = OrderedDict(data_out)
        elif self.owner is not None:
            if isinstance(postsynch, str):
                postsynch = [postsynch]
            if '*' in postsynch:
                watched = self.__get_owner_connections()
                varnames = watched.keys()
            else:
                owner = self.owner
                def writeable(var) -> bool:
                    try:
                        owner[var] = owner[var]
                    except AttributeError:
                        return False
                    else:
                        return True
                matches = find_variables(
                    owner,
                    includes=postsynch,
                    excludes=None,
                    inputs=False,
                )
                varnames = set(filter(writeable, matches.keys()))
                # Make sure varlist contains at least owner system outputs
                for portname, port in owner.outputs.items():
                    varnames.update(natural_varname(f"{portname}.{var}") for var in port)
            doe_out = OrderedDict((name, []) for name in varnames)
        return doe_out

    def __get_doe_out_sizes(self) -> None:
        state = self.__state
        owner = self.__owner
        doe_out_sizes = OrderedDict.fromkeys(state.doe_out.keys(), None)
        pos = 0
        for var in doe_out_sizes:
            size = 0
            attr = getattr(owner, var)
            if isinstance(attr, (Number, numpy.ndarray)):
                size = numpy.size(attr)
            else:
                raise TypeError(
                    f"Unsupported data type {type(attr).__name__} for {owner.name}.{var}"
                )
            doe_out_sizes[var] = (pos, size)
            pos += size
        self.__state = SystemSurrogateState(
            state.doe_in,
            state.doe_out,
            state.model,
            doe_out_sizes,
        )

    def __format_outputs(self) -> numpy.ndarray:
        logger.debug(f"Reshaping outputs")
        state = self.__state
        reshaped_outputs = []
        for k in range(len(state.doe_in)):
            reshaped_outputs.append(
                list(value[k] for value in state.doe_out.values())
            )
        reshaped_outputs = numpy.asarray(list(list(flatten(el)) for el in reshaped_outputs))
        # logger.debug(f"Reshaped outputs are now numpy.ndarray and are: {reshaped_outputs}")
        return reshaped_outputs

    def __format_inputs(self) -> numpy.ndarray:
        logger.debug(f"Reshaping input data")
        res = []
        for couple_values in self.__state.doe_in.values:
            res.append(list(flatten(couple_values)))
        return numpy.asarray(res)

    def __set_and_execute(self) -> None:
        logger.debug(f"Setting and executing in order to build data for training")
        owner = self.__owner
        state = self.__state
        for i, row in state.doe_in.iterrows():
            logger.log(LogLevel.FULL_DEBUG, f"Setting {owner.name!r} input values (DOE row #{i})")
            for var, value in row.items():
                owner[var] = value
            owner.run_drivers()
            # add tracked output data to doe_out
            for var in state.doe_out:
                state.doe_out[var].append(owner[var])

    def add_data(self, newdoe: pandas.DataFrame) -> NoReturn:
        # TODO: Should merge input lists when matching names (not working yet)
        # self.__state = SystemSurrogateState(
        #     self.__state.doe_in,
        #     self.__state.doe_out,
        #     self.__state.model,
        #     OrderedDict(),
        # )
        raise NotImplementedError()

    def __prepare_and_train(self) -> None:
        logger.debug(f"Preparing and training function")
        if len(self.__state.doe_out) == 0:
            raise ValueError(
                f"Cannot train surrogate model: no output found in System {self.owner.name!r}"
            )
        self.__get_doe_out_sizes()
        if self.__need_doe:
            self.__set_and_execute()
        x = self.__format_inputs()
        y = self.__format_outputs()
        self.__train_model(x, y)

    def __train_model(self, x, y) -> None:
        self.__state.model.train(x, y)

    def predict(self, x):
        return self.__state.model.predict(x).reshape(1, -1)[0]

    def compute(self) -> None:
        logger.debug(f"Start meta_compute() instead of {self.owner.name}.compute().")
        inputs = self.__get_owner_inputs()
        outputs = self.predict(inputs)
        self.__set_owner_outputs(outputs)
        logger.debug(f"meta_compute() done")

    def __get_owner_inputs(self) -> numpy.ndarray:
        owner = self.owner
        logger.debug(f"Getting current inputs of system {owner.name}")
        inputs = list(flatten(owner[var] for var in self.__state.doe_in.columns))
        logger.debug(f"Detected input values: {inputs}")
        return numpy.asarray(inputs)

    def __set_owner_outputs(self, np_outputs: numpy.ndarray) -> None:
        owner = self.__owner
        logger.log(LogLevel.FULL_DEBUG, f"Post-synchronize outputs of {owner.name}")
        for var, (pos, size) in self.__state.doe_out_sizes.items():
            attr = getattr(owner, var)
            if isinstance(attr, numpy.ndarray):
                attr.flat = np_outputs[pos : pos + size]
            elif size == 1:
                setattr(owner, var, np_outputs[pos])
            else:
                # Should never occur, according to `__get_doe_out_sizes`
                raise TypeError(
                    f"Unsupported data type {type(attr).__name__} for {owner.name}.{var}"
                )

    def dump(self, filename: str) -> None:
        """Dump current state to file"""
        logger.debug(f"Dumping metamodel in {filename}")
        with open(filename, 'wb') as fp:
            pickle.dump(self.__state, fp)

    @classmethod
    def load(cls, owner: "cosapp.systems.System", filename: str) -> "SystemSurrogate":
        """Load system surrogate from file"""
        logger.debug(f"Loading metamodel from {filename}")
        from cosapp.systems import System
        check_arg(owner, 'owner', System)
        with open(filename, 'rb') as fp:
            state = pickle.load(fp)
        if not isinstance(state, SystemSurrogateState):
            raise TypeError("cannot load object")
        obj = cls(owner, data_in={}, model=None)
        obj.__state = state
        return obj

    def __check_unknowns_and_transients(self) -> None:
        state = self.__state
        unknowns = get_unknowns_transients(self.__owner)
        unsolvable_unknowns = set(unknowns).difference(state.doe_out)
        trained_unknowns = unsolvable_unknowns.issubset(state.doe_in)
        if not trained_unknowns:
            warnings.warn(
                "The following unknowns/transients are not part of the training set; "
                f"future attempts to compute them with a driver may fail: "
                f"{list(unknowns.difference(state.doe_in))}"
            )

    def __get_owner_connections(self) -> "OrderedDict[str, list]":
        return get_dependent_connections(self.__owner)


def flatten(iterable: Iterable) -> Iterable:
    for elem in iterable:
        if isinstance(elem, collections.abc.Iterable) and not isinstance(elem, str): 
            yield from flatten(elem)
        else:
            yield elem


def get_dependent_connections(system: "cosapp.systems.System") -> Dict[str, PortType]:
    """This function returns a dictionnary mapping variable names to a port direction.
    Keys are absolute paths to connected inputs and all outputs.
    Values are owner port direction.
    """
    from cosapp.base import System

    def get_connections(system: System, head_system: System) -> Dict[str, PortType]:
        """Recursive inner version of `get_dependent_connections`"""
        result = dict()
        prefix = ""
        if system is not head_system:
            prefix = f"{head_system.get_path_to_child(system)}."
            for connector in system.incoming_connectors():
                sink = connector.sink
                logger.debug(f"Detecting connector {connector} with sink {sink.name!r}")
                if sink.is_input:
                    for var in connector.sink_variables():
                        key = f"{prefix}{sink.name}.{var}"
                        result[key] = PortType.IN
                        logger.debug(f"Add {key} to list of connected inputs")

        for output in system.outputs.values():
            logger.debug(f"Checking output {output}.")
            for var in output:
                key = f"{prefix}{output.name}.{var}"
                result[key] = PortType.OUT
                logger.debug(f"Add {key} to list of outputs")
        if system.is_standalone():
            unknowns = system.assembled_problem().unknowns
            for unknown in unknowns:
                key = f"{prefix}{unknown}"
                result[key] = PortType.IN
                logger.debug(f"Add {key} to list of unknowns")
        for child in system.children.values():
            logger.debug(
                f"Targetted child of recursive getter of connected inputs and all outputs is {child}"
            )
            result.update(get_connections(child, head_system))
        return result

    logger.debug(
        f"Recursive search of connected inputs and all outputs on {system.name}"
    )
    return get_connections(system, system)


def get_unknowns_transients(system: "cosapp.systems.System") -> set:
    res = set()
    problem = system.assembled_problem()
    for collection in (problem.unknowns, problem.transients):
        res.update(collection)
    return res



import enum
import re
from typing import Dict, NamedTuple
from cosapp.utils.helpers import check_arg


@enum.unique
class EventDirection(enum.Enum):
    """Enum covering zero-crossing directions"""
    UP = {
        'desc': "Upward zero-crossing",
        'func': (lambda prev, curr: curr > prev and curr * prev <= 0),
    }
    DOWN = {
        'desc': "Downward zero-crossing",
        'func': (lambda prev, curr: curr < prev and curr * prev <= 0),
    }
    UPDOWN = {
        'desc': "Up- or downward zero-crossing",
        'func': (lambda prev, curr: curr != prev and prev * curr <= 0),
    }

    def zero_detected(self, prev, curr) -> bool:
        detector = self.value['func']
        return detector(prev, curr)


class ZeroCrossing(NamedTuple):
    expression: str
    direction: EventDirection

    @classmethod
    def up(cls, expression: str) -> "ZeroCrossing":
        return cls(expression, EventDirection.UP)
 
    @classmethod
    def down(cls, expression: str) -> "ZeroCrossing":
        return cls(expression, EventDirection.DOWN)
 
    @classmethod
    def updown(cls, expression: str) -> "ZeroCrossing":
        return cls(expression, EventDirection.UPDOWN)
 
    @classmethod
    def from_comparison(cls, expression: str) -> "ZeroCrossing":
        """Interpret an expression of the kind 'lhs <op> rhs'
        as a `ZeroCrossing` object, where <op> is one of
        comparison operators:
        - `<`, `<=`
        - `==`
        - `>`, `>=`
        """
        check_arg(expression, "expression", str)
        operators = cls.operators()
        not_in_sides = list(operators) + ["="]

        def side_ok(side: str) -> bool:
            return not any(nope in side for nope in not_in_sides) and side.strip()

        for operator, direction in operators.items():
            try:
                lhs, rhs = expression.split(operator, maxsplit=1)
            except:
                continue
            if side_ok(lhs) and side_ok(rhs):
                return cls(f"{lhs.strip()} - ({rhs.strip()})", direction)

        raise ValueError(
            f"Expected a comparison between two expressions involving one of {tuple(operators)}"
            f"; got {expression!r}."
        )

    @classmethod
    def operators(cls) -> Dict[str, EventDirection]:
        return {
            "<" : EventDirection.DOWN,
            "<=": EventDirection.DOWN,
            "==": EventDirection.UPDOWN,
            ">=": EventDirection.UP,
            ">" : EventDirection.UP,
        }



import numpy
from typing import Any, Optional, Tuple, List, NamedTuple
from collections.abc import Collection

from .helpers import check_arg
from .naming import natural_varname


class MaskedVarInfo(NamedTuple):
    basename: str
    selector: str = ""
    mask: Optional[numpy.ndarray] = None

    @property
    def fullname(self) -> bool:
        return f"{self.basename}{self.selector}"


def find_selector(expression: str) -> Tuple[str, str]:
    """Decompose a string expression into `basename` and `selector`,
    where `selector` is a suitable mask expression for an array.

    Parameters:
    -----------
    expression [str]: the expression to be parsed.
    
    Returns:
    --------
    baseline, selector [str, str]
    """
    expression = expression.strip()
    left_bracket, right_bracket = brackets = tuple("[]")
    nl = nr = il = 0

    basename, selector = expression, ""  # default

    if expression.endswith(right_bracket):
        i = len(expression)
        for char in reversed(expression):
            i -= 1
            if char == right_bracket:
                nr += 1
            elif char == left_bracket:
                nl += 1
                if nl == nr:
                    prev_char = expression[i - 1] if i > 0 else None
                    if prev_char not in brackets:
                        il = i
                        break
    
    elif expression.endswith(left_bracket):
        nl, nr = 1, 0
    
    balanced = (nl == nr)
    if not balanced:
        raise ValueError(f"Bracket mismatch in {expression!s}")
    if il > 0 or expression.startswith(left_bracket):
        basename = expression[:il]
        selector = expression[il:]

    return basename, selector


# TODO should we reintegrate it as Boundary method as it is now the only used place
def get_indices(system: "cosapp.base.System", name: str) -> MaskedVarInfo:
    """Decompose a variable specification into its base name and mask.

    Parameters
    ----------
    system : System
        System to which variable belongs
    name : str
        Variable specification (variable name + optional array mask, if required)

    Returns
    -------
    MaskedVarInfo (named tuple):
        - basename [str]: variable name
        - selector [str]: array selector
        - mask [numpy.ndarray[bool]]: mask (if array) or `None`
    """
    check_arg(name, 'name', str)
    name = natural_varname(name)

    def check_eval(attribute: str) -> Any:
        try:
            value = eval(f"s.{attribute}", {}, {"s": system})
        except AttributeError as error:
            error.args = (f"{attribute!r} is not known in {system.name}",)
            raise
        except Exception as error:
            error.args = (f"Can't evaluate {attribute!r} in {system.name}",)
            raise
        else:
            return value

    try:
        basename, selector = find_selector(name)
    except ValueError as error:
        raise SyntaxError(error)

    value = check_eval(basename)
    mask = None

    if selector:
        # Check value is an array
        if not (
            isinstance(value, numpy.ndarray)
            and numpy.issubdtype(value.dtype, numpy.number)
            and value.size > 1
        ):
            raise TypeError(
                f"Only non-empty numpy arrays can be partially selected; got '{system.name}.{name}'."
            )
        mask = numpy.zeros_like(value, dtype=bool)
        # Set mask from selector string
        try:
            exec(f"mask{selector} = True", {}, {"mask": mask})
        except (SyntaxError, IndexError) as error:
            varname = f"{system.name}.{basename}"
            error.args = (
                f"Invalid selector {selector!r} for variable {varname!r}: {error!s}",
            )
            raise
    
    elif isinstance(value, Collection) and not isinstance(value, str):
        mask = numpy.ones_like(value, dtype=bool)

    return MaskedVarInfo(basename, selector, mask)


def multi_split(expression: str, separators: List[str]) -> Tuple[List[str], List[str]]:
    """Extension of `str.split`, accounting for more than one split separators.
    
    Parameters:
    -----------
    - expression [str]:
        Expression to be split.
    - separators [List[str]]:
        List of separators.
    
    Returns:
    --------
    - expressions [List[str]]:
        List of n split expressions.
    - separators [List[str]]:
        Sequence of (n - 1) separators between split expressions.

    Examples:
    ---------
    >>> multi_split('a+b-c-d+e', list('+-'))
    ['a', 'b', 'c', 'd', 'e'], ['+', '-', '-', '+']
    """
    expressions = [expression]
    sep_list = []

    for separator in set(separators):
        new_list = []
        shift = 0
        for i, expression in enumerate(expressions):
            sublist = expression.split(separator)
            n_hits = len(sublist) - 1
            new_list.extend(sublist)
            if n_hits > 0:
                j = i + shift
                shift += n_hits
                sep_list = sep_list[:j] + [separator] * n_hits + sep_list[j:]
        expressions = list(map(str.strip, new_list))

    return expressions, sep_list



"""Integration tests for time drivers
"""
import pytest
import numpy as np

from cosapp.base import System
from cosapp.tests.library.systems import BouncingBall, PointMass
from cosapp.drivers import NonLinearSolver, RunSingleCase, RungeKutta
from cosapp.recorders import DataFrameRecorder
from typing import Tuple


@pytest.fixture
def ball() -> BouncingBall:
    ball = BouncingBall('ball', g=[0, -9.81])
    ball.mass = 1.5
    ball.cf = 0.2
    ball.cr = 0.98
    return ball


BallCase = Tuple[BouncingBall, RungeKutta]

@pytest.fixture
def ball_case(ball: BouncingBall) -> BallCase:
    """Bouncing ball test case.
    Returns tuple (system, time_driver).
    """
    driver = ball.add_driver(RungeKutta(order=3, dt=0.01))
    driver.time_interval = (0, 20)

    # Add a recorder to capture time evolution in a dataframe
    driver.add_recorder(
        DataFrameRecorder(
            includes = ['x', 'v', 'a', 'n_r*', 'norm(v)'],
        ),
        period = 0.05,
    )
    # Define a simulation scenario
    driver.set_scenario(
        init = {
            'x': np.array([0, 0]),
            'v': np.array([8, 9.5]),
        },
        stop = ball.rebound.filter("norm(v) < 1"),
        # stop = ball.rebound.filter("n_rebounds == 3"),
    )
    return ball, driver


def test_BouncingBall_traj(ball_case: BallCase):
    ball, driver = ball_case
    driver.scenario.stop.trigger = ball.rebound

    ball.run_drivers()

    assert ball.x == pytest.approx([6.416270, 0])
    # Retrieve recorded data
    # data = driver.recorder.export_data()
    # data = data.drop(['Section', 'Status', 'Error code'], axis=1)
    # print("", data, sep="\n")
    # time = np.asarray(data['time'])
    # traj = np.asarray(data['x'].tolist())


def test_BouncingBall_design_cf():
    """Test involving a time driver nested in a nonlinear solver,
    with a system containing a primary event. If events are not
    properly reset at the beginning of each time simulation, the solver
    fails at the first iteration, as the internal state of the event
    is inconsistent.
    Related to https://gitlab.com/cosapp/cosapp/-/issues/90
    """
    class BallDesignCase(System):
        def setup(self):
            ball = self.add_child(BouncingBall('ball', g=[0, -9.81]))
            self.add_event('first_rebound', trigger=ball.rebound.filter('n_rebounds == 0'))
            self.add_event('reach_top', trigger='ball.v[-1] == 0')
            self.add_outward_modevar('t_rebound', init=0.0)
    
        def transition(self) -> None:
            if self.first_rebound.present:
                self.t_rebound = self.time

    s = BallDesignCase('s')
    ball = s.ball
    ball.mass = 1.0
    ball.cf = 0.1
    ball.cr = 1.0

    solver = s.add_driver(NonLinearSolver('solver', tol=1e-9))
    solver.add_driver(RunSingleCase('runner'))
    driver = solver.runner.add_driver(
        RungeKutta(dt=0.05, time_interval=[0, 2], order=2)
    )
    driver.set_scenario(
        init = {
            'ball.x': np.array([0, 0]),
            'ball.v': np.array([8, 9.5]),
        },
        stop = ball.rebound.filter("norm(v) < 0.1"),
    )

    # Setup design problem
    solver.add_unknown('ball.cf', max_rel_step=0.5).add_equation('t_rebound == 1.522')

    # Add recorders, for debug
    # solver.runner.add_recorder(
    #     DataFrameRecorder(includes=['ball.cf', 't_rebound'], hold=True)
    # )
    # driver.add_recorder(
    #     DataFrameRecorder(
    #         includes=[f"ball.{var}" for var in 'x'],
    #         hold=False,
    #     ),
    #     period=0.05,
    # )

    s.ball.cf = 0.5  # first guess
    s.run_drivers()

    # Retrieve recorded data
    # if solver.runner.recorder and driver.recorder:
    #     res = solver.runner.recorder.export_data()
    #     data = driver.recorder.export_data()
    #     data = data.drop(['Section', 'Status', 'Error code'], axis=1)
    #     print("", res, data, sep="\n")

    assert s.ball.cf == pytest.approx(0.1045, rel=1e-3)



import pytest

from cosapp.base import System
from cosapp.drivers import EulerExplicit
from cosapp.recorders import DataFrameRecorder


class ThresholdSystem(System):
    def setup(self):
        self.add_inward('x', 0.0)
        self.add_inward('threshold', 0.0)
        self.add_outward_modevar('below', True, init='x < threshold')
        self.add_event('cross', trigger='x == threshold')
    
    def transition(self):
        if self.cross.present:
            self.below = not self.below


def test_ModeVariableIntegration():
    s = ThresholdSystem('s')
    driver = s.add_driver(
        EulerExplicit(dt=0.1, time_interval=(0, 2))
    )
    driver.set_scenario(
        values = {
            'threshold': 0.6,
            'x': 'cos(4 * t)',
        },
    )
    driver.add_recorder(
        DataFrameRecorder(includes=['x', 'below', 'x - threshold']),
        period = 0.1
    )
    s.below = True  # inconsistent with `init` expression
    s.run_drivers()

    data = driver.recorder.export_data()
    data = data.drop(['Section', 'Status', 'Error code'], axis=1)
    # print('\n', data)
    assert not data['below'][0]  # consistent with `init` expression
    
    threshold = s.threshold

    for i, row in data.iterrows():
        x = row['x']
        below = row['below']
        delta = x - threshold
        context = f"row #{i}, x = {x}"
        if delta < -1e-12:
            assert below, context
        elif delta > 1e-12:
            assert not below, context



import io
import abc
from typing import Dict
from jinja2.environment import Template

from cosapp.systems import System
from cosapp.utils.helpers import check_arg


class BaseRenderer(metaclass=abc.ABCMeta):
    """
    Base class to export a system as HTML.

    Parameters
    ----------
    system : System
        System to export
    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    def __init__(self, system: System, embeddable=False):
        check_arg(system, "system", System)
        self.system = system
        self.embeddable = embeddable

    @abc.abstractmethod
    def get_data(self, **kwarg) -> Dict:
        """Convert  `self.system` into a dictionary used to build the HTML page.

        Returns
        -------
        Dict
            Dictionary containing elements to create HTML page
        """
        pass

    @abc.abstractmethod
    def html_content(self) -> str:
        """Returns HTML content of renderer's system as a character string."""
        pass

    def dump(self, fstream: io.RawIOBase) -> None:
        """Dump HTML content into writable fstream"""
        rendered_html = self.html_content()
        fstream.write(rendered_html)

    def to_file(self, filename: str) -> None:
        """Dump HTML content into text file `filename`"""
        with open(filename, "w", encoding="utf-8") as fp:
            self.dump(fp)

    @classmethod
    def html_tags(cls) -> Dict[str, str]:
        html_begin_tags = """<!DOCTYPE html>
            <html>
            <head>
            <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            <title>{{title}}</title>
            </head>
            <body>\n
        """
        html_end_tags = """
            </body>
            </html>
        """
        return {
            "html_begin_tags": html_begin_tags,
            "html_end_tags": html_end_tags,
        }

    @classmethod
    @abc.abstractclassmethod
    def html_resources(cls) -> Dict[str, str]:
        """
        Return the necessary resources to render Jinja template.
        """
        pass

    @classmethod
    @abc.abstractclassmethod
    def html_template(cls) -> Template:
        """
        Return the Jinja template used to create HTML file.
        """
        pass

    @classmethod
    def get_globals(cls) -> Dict:
        """Returns a dict containing class-wide environment and HTML data"""
        return dict(
            template=cls.html_template(),
            **cls.html_tags(),
            **cls.html_resources(),
        )



from fnmatch import fnmatchcase
from typing import List, Any, Callable, Set, Dict, Union
from collections.abc import Collection
import inspect
import itertools

from cosapp.ports.port import BasePort, Port
from cosapp.utils.naming import natural_varname, CommonPorts
from cosapp.utils.helpers import check_arg

SearchPattern = Union[str, List[str]]


def make_wishlist(wishlist: SearchPattern, name="wishlist") -> List[str]:
    ok = True
    if isinstance(wishlist, str):
        wishlist = [wishlist]
    elif isinstance(wishlist, Collection):
        ok = len(wishlist) == 0 or all(isinstance(item, str) for item in wishlist)
    elif wishlist is None:
        wishlist = []
    else:
        ok = False
    if not ok:
        raise TypeError(
            f"{name!r} must be a string, or a sequence of strings; got {wishlist}."
        )
    # Filter out common port names 'inwards.', 'outwards.',.. from wishlist
    filtered = set(map(natural_varname, wishlist))
    return list(filtered)


def get_attributes(obj) -> Set[str]:
    """Returns non-callable members of an object"""
    return set(
        m[0] for m in inspect.getmembers(obj)
        if not m[0].startswith("_") and not callable(m[1])
    )


def find_variables(
    system: "cosapp.systems.System",
    includes: SearchPattern,
    excludes: SearchPattern,
    advanced_filter: Callable[[Any], bool] = lambda any: True,
    inputs: bool = True,
    outputs: bool = True,
    include_const: bool = False,
) -> Dict[str, Any]:
    """Generate the dictionary (name, value) of variables
    whose names match inclusion and exclusion criteria.

    Parameters
    ----------
    system : cosapp.systems.System
        Object that owns the variables searched.
    includes : str or List[str]
        Variables matching these patterns will be included.
    excludes : str or List[str]
        Variables matching these patterns will be excluded.
    advanced_filter : Callable[[Any], bool]
        Function taking the variable as input and returning an acceptance criteria (True if variable is valid).
    inputs : bool
        Defines if input variables will be accepted or not.
    outputs : bool
        Defines if output variables will be accepted or not.
    include_const : bool
        Defines if read-only properties defined with `System.add_property` will be accepted or not.

    Returns
    -------
    dict[str, Any]
        Dictionary (name, value) of matches.

    .. note::
        Inward and outward variables will appear without the prefix `inwards.` or `outwards.`.
    """
    from cosapp.systems import System  # Local import to avoid recursion
    check_arg(system, 'system', System)

    if not (inputs or outputs):
        # quick return, if possible
        return []

    includes = make_wishlist(includes)
    excludes = make_wishlist(excludes)
    result = dict()

    def is_valid(port: BasePort) -> bool:
        return isinstance(port, BasePort) and (
            (port.is_input and inputs) or
            (port.is_output and outputs)
        )

    def find_matches(name: str, value: Any) -> Dict[str, Any]:
        matches = dict()
        if advanced_filter(value):
            for pattern in includes:
                if fnmatchcase(name, pattern):
                    include = True
                    for pattern in excludes:
                        if fnmatchcase(name, pattern):
                            include = False
                            break
                    if include:
                        matches[name] = value
                        break
        return matches

    ports = set()

    for name, ref in system.name2variable.items():
        # Save variable for non virtual port
        # Suppress duplicates INWARDS and OUTWARDS
        port = ref.mapping
        valid = (
            is_valid(port)
            and name == natural_varname(name)
        )
        if not valid:
            continue  # skip `name`

        ports.add(port)
        result.update(find_matches(name, ref.value))
    
    # Find matches among port properties
    Port_cls_attr = get_attributes(Port)

    for port in ports:
        if not isinstance(port, Port):
            continue  # exclude `inwards` and `outwards` extensible ports
        port_properties = get_attributes(port) - Port_cls_attr - set(port._variables)
        port_name = port.full_name(trim_root=True)
        for name in port_properties:
            result.update(
                find_matches(f"{port_name}.{name}", getattr(port, name))
            )

    # Find matches among system properties
    system_properties = find_system_properties(system, include_const)
    for name in system_properties:
        try:
            child, attr = name.rsplit('.', maxsplit=1)
        except ValueError:
            owner, attr = system, name
        else:
            owner = system[child]
        result.update(
            find_matches(name, getattr(owner, attr))
        )

    return result


def find_variable_names(
    system: "cosapp.systems.System",
    includes: SearchPattern,
    excludes: SearchPattern,
    advanced_filter: Callable[[Any], bool] = lambda any: True,
    inputs: bool = True,
    outputs: bool = True,
    include_const: bool = False,
) -> List[str]:
    """Generate the list of requested variable names,
    given inclusion and exclusion criteria.

    Matching variable names are returned in alphabetical order.

    Parameters
    ----------
    system : cosapp.systems.System
        Object that owns the variables searched.
    includes : str or List[str]
        Variables matching these patterns will be included.
    excludes : str or List[str]
        Variables matching these patterns will be excluded.
    advanced_filter : Callable[[Any], bool]
        Function taking the variable as input and returning an acceptance criteria (True if variable is valid).
    inputs : bool
        Defines if input variables will be accepted or not.
    outputs : bool
        Defines if output variables will be accepted or not.
    include_const : bool
        Defines if read-only properties defined with `System.add_property` will be accepted or not.

    Returns
    -------
    list[str]
        Variable names in `system` matching the requested includes/excludes patterns.

    .. note::
        Inward and outward variables will appear without the prefix `inwards.` or `outwards.`.
        This functions returns the sorted keys of the dictionary returned by `find_variables`.
    """
    return sorted(
        find_variables(
            system,
            includes,
            excludes,
            advanced_filter,
            inputs,
            outputs,
            include_const,
        )
    )


def find_system_properties(system, include_const=False) -> Set[str]:
    """
    Returns system properties, defined either as class properties
    (with @property decorator), or with `System.add_property`.
    The latter are excluded if optional argument `include_const`
    is False (default).

    Parameters
    ----------
    - system [cosapp.systems.System]:
        System of interest
    - include_const [bool, optional]:
        Defines if read-only properties defined with `System.add_property` will be accepted or not..
        Default: `False`.

    Returns
    -------
    - set[str]:
        Set of property names.
    """
    from cosapp.systems import System  # Local import to avoid recursion
    check_arg(system, 'system', System)

    base_cls_attr = get_attributes(System)
    get_name = lambda obj: obj.name

    def local_properties(system: System) -> Set[str]:
        local_attr = get_attributes(system) - base_cls_attr
        writable = set(
            map(get_name,
                itertools.chain(
                    system.ports(),
                    system.events(),
                    system.children.values(),
                )
            )
        )
        local_attr -= writable
        for port_name in CommonPorts.names():
            local_attr -= set(system[port_name])
        if not include_const:
            local_attr -= set(system.properties)
        return local_attr

    tree = system.tree(downwards=True)
    properties = local_properties(next(tree))
    for child in tree:
        prefix = f"{system.get_path_to_child(child)}."
        properties |= set(
            f"{prefix}{attr}"
            for attr in local_properties(child)
        )
    
    return properties



"""Test export to FMU"""
import importlib
import itertools
import re
import uuid
import zipfile

import numpy
import pandas
import pytest
from numpy.testing import assert_almost_equal

from cosapp.drivers import MonteCarlo, NonLinearSolver, RungeKutta, RunSingleCase
from cosapp.ports.enum import PortType
from cosapp.ports.port import ExtensiblePort
from cosapp.recorders import DataFrameRecorder
from cosapp.systems import System
from cosapp.tools.fmu.exporter import (
    Fmi2Causality,
    Fmi2Variability,
    FmuBuilder,
    TimeIntegrator,
    VariableType,
    to_fmu,
)

# Check optional third parties availability
pytest.importorskip("jinja2")
pytest.importorskip("pythonfmu")


def load_fmu_class(module_name: str, filepath: str):
    # Import the user interface
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    fmu_interface = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(fmu_interface)
    # Instantiate the interface
    with open(filepath, "r") as f:
        data = f.read()
        class_name = re.search("class (\w+)\(\s*Fmi2Slave\s*\)\s*:", data).group(1)
    return getattr(fmu_interface, class_name)(instance_name="instance")


def get_variable_names(vars):
    return [var.name for var in vars]


@pytest.mark.parametrize("value", TimeIntegrator)
def test_TimeIntegrator_driver(value):
    assert (
        value.driver == "RungeKutta"
        if value.value.startswith("Runge-Kutta")
        else value.name
    )


@pytest.mark.parametrize("value", TimeIntegrator)
def test_TimeIntegrator_options(value):
    expected = (
        dict(order=int(value.value.split()[1]))
        if value.value.startswith("Runge-Kutta")
        else dict()
    )
    assert value.options == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (True, VariableType.Boolean),
        (2, VariableType.Integer),
        (42.0, VariableType.Real),
        ("hello", VariableType.String),
        (b"hello", TypeError),
        (object(), TypeError),
        (numpy.array(True), VariableType.Boolean),
        (numpy.array(2), VariableType.Integer),
        (numpy.array(42.0), VariableType.Real),
        (numpy.array("hello", dtype="U"), VariableType.String),
        (numpy.array([1, 2]), TypeError),
        (numpy.array(b"hello", dtype="S"), TypeError),
    ],
)
def test_FMUBuilder__get_variable_type(value, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            FmuBuilder._get_variable_type(value)
    else:
        assert FmuBuilder._get_variable_type(value) == expected


@pytest.mark.parametrize(
    "causality, variability",
    itertools.product(Fmi2Causality, (list(Fmi2Variability) + [None])),
)
def test_FMUBuilder__add_variables(causality, variability):
    data = dict((("a", True), ("b", 2), ("c", 42.0), ("d", "hello")))

    vars = FmuBuilder._add_variables(data, causality, variability)

    for v in vars:
        assert v.name in data
        assert v.causality == causality.name
        if variability is None:
            assert v.variability is None
        else:
            assert v.variability == variability.name
    assert len(vars) == len(data)

    with pytest.raises(TypeError):
        FmuBuilder._add_variables({"data": b"hello"}, causality, variability)


def test_FMUBuilder__get_default_value(testtype):
    variables = FmuBuilder._get_default_value(
        list(testtype.inputs[System.INWARDS]), testtype
    )

    for k, v in variables.items():
        if k == "k":
            assert v == pytest.approx(getattr(testtype, k))
        else:
            assert v == getattr(testtype, k)

    variables = FmuBuilder._get_default_value(("k[0]", "k[1]", "k[2]"), testtype)

    for i in range(3):
        assert variables[f"k[{i}]"] == testtype["k"][i]

    with pytest.raises(ValueError):
        FmuBuilder._get_default_value(("k[4]",), testtype)

    with pytest.raises(ValueError):
        FmuBuilder._get_default_value(("rr",), testtype)


def test_FMUBuilder__get_default_variables():
    p1 = ExtensiblePort(System.INWARDS, PortType.IN)
    p1.add_variable("a", True)
    p1.add_variable("b", 2)
    p1.add_variable("c", 42.0)
    p1.add_variable("d", "hello")
    p1.add_variable("e", b"hello")

    p2 = ExtensiblePort("out", PortType.OUT)
    p2.add_variable("f", object())
    p2.add_variable("g", numpy.array(True))
    p2.add_variable("h", numpy.array(2))
    p2.add_variable("i", numpy.array(42.0))
    p2.add_variable("j", numpy.array("hello", dtype="U"))
    p2.add_variable("k", numpy.array([1, 2]))
    p2.add_variable("l", numpy.array(b"hello", dtype="S"))

    out = FmuBuilder._get_default_variables({"p1": p1, System.OUTWARDS: p2}, to_skip=())
    expected = ("p1.a", "p1.b", "p1.c", "p1.d", "g", "h", "i", "j")
    for e in expected:
        assert e in out
    assert len(out) == len(expected)

    out = FmuBuilder._get_default_variables(
        {"p1": p1, System.OUTWARDS: p2}, to_skip=["p1.c", "p1.a", "j"]
    )
    expected = ("p1.b", "p1.d", "g", "h", "i")
    for e in expected:
        assert e in out
    assert len(out) == len(expected)


def test_FMU_export_all_causality(tmp_path, allcausality):
    test = allcausality
    fmu_file = to_fmu(test, dest=tmp_path)

    if fmu_file.suffix == ".fmu":
        with zipfile.ZipFile(fmu_file, "r") as zip_desc:
            zip_desc.extractall(fmu_file.parent)

    fmu_slave = load_fmu_class(
        "allcausality", str(fmu_file.parent / "resources" / "allcausality.py")
    )

    assert len(fmu_slave.vars) == 4
    names = get_variable_names(fmu_slave.vars.values())
    assert "in_.x" in names
    assert fmu_slave.vars[names.index("in_.x")].causality == Fmi2Causality.input
    assert "in_x" in names
    assert fmu_slave.vars[names.index("in_x")].causality == Fmi2Causality.parameter
    assert "out.x" in names
    assert fmu_slave.vars[names.index("out.x")].causality == Fmi2Causality.output
    assert "out_y" in names
    assert fmu_slave.vars[names.index("out_y")].causality == Fmi2Causality.local


def test_FMU_export_hide_all_variables(tmp_path, allcausality):
    test = allcausality
    fmu_folder = FmuBuilder.generate_fmu_facade(
        test, dest=tmp_path, inputs=[], locals=[], outputs=[], parameters=[]
    )

    fmu_slave = load_fmu_class(
        "allcausality",
        str(FmuBuilder._get_project_folder(tmp_path) / "allcausality.py"),
    )

    assert len(fmu_slave.vars) == 0


def test_FMU_export_all_type(tmp_path, testtype):
    test = testtype
    fmu_folder = FmuBuilder.generate_fmu_facade(test, dest=tmp_path)

    fmu_slave = load_fmu_class(
        "testtype", str(FmuBuilder._get_project_folder(tmp_path) / "testtype.py")
    )

    assert len(fmu_slave.vars) == 8


@pytest.mark.parametrize("name", ["k", "l"])
def test_FMU_export_type_failure(tmp_path, testtype, name):
    test = testtype
    with pytest.raises(TypeError):
        to_fmu(test, dest=tmp_path, parameters={name: testtype[name]})


def test_FMU_export_non_linear_multipoints(tmp_path, ode):
    solver = NonLinearSolver("solver")
    solver.add_child(RunSingleCase("pt1"))
    solver.add_child(RunSingleCase("pt2"))

    with pytest.raises(ValueError):
        to_fmu(ode, dest=tmp_path, nonlinear_solver=solver)


def test_FMU_export_non_linear_bad_child(tmp_path, ode):
    solver = NonLinearSolver("solver")
    solver.add_child(MonteCarlo("dummy"))

    with pytest.raises(TypeError):
        to_fmu(ode, dest=tmp_path, nonlinear_solver=solver)


def test_FMU_handle_vector_variable(tmp_path, vector_syst):
    pyfmi = pytest.importorskip("pyfmi")

    fmu_file = to_fmu(
        vector_syst,
        dest=tmp_path,
        inputs=["vi[0]", "vr[1]"],
        outputs=["vi[2]", "vr[2]"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    variables = model.get_model_variables()

    assert model.get_integer([variables["vi[0]"].value_reference])[0] == 1
    model.set_integer([variables["vi[2]"].value_reference], [33])
    assert model.get_integer([variables["vi[2]"].value_reference])[0] == 33

    assert model.get_real([variables["vr[2]"].value_reference])[0] == 3.0
    model.set_real([variables["vr[1]"].value_reference], [42.0])
    assert model.get_real([variables["vr[1]"].value_reference])[0] == 42.0



@pytest.mark.parametrize("integrator", TimeIntegrator)
def test_FMU_export_time_integrator(tmp_path, ode, integrator):
    pyfmi = pytest.importorskip("pyfmi")
    fmu_file = to_fmu(
        ode,
        dest=tmp_path,
        locals={"f": 0.0, "df_dt": 0.0},
        time_integrator=integrator,
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    results = model.simulate(final_time=1)

    # Reference value
    end_time = results["time"][-1]
    ref = ode(end_time)
    assert results["f"][-1] == pytest.approx(ref, rel=1e-3)



def test_FMU_integration_simple(tmp_path, ode):
    pyfmi = pytest.importorskip("pyfmi")

    # Convert to FMU
    fmu_file = to_fmu(
        ode,
        dest=tmp_path,
        locals=["f", "df_dt"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    results = model.simulate(final_time=1)

    # Reference value
    end_time = results["time"][-1]
    ref = ode(end_time)
    assert results["f"][-1] == pytest.approx(ref, rel=1e-7)


def test_FMU_integration_nonlinear_design(tmp_path, iterativenonlinear):
    """Same as `test_FMU_integration_nonlinear_local` with additional
    problem defined at solver level (off-design problem).
    """
    pyfmi = pytest.importorskip("pyfmi")

    # Init
    t_driver = iterativenonlinear.add_driver(
        RungeKutta(order=4, dt=0.1, time_interval=[0, 0.2])
    )
    solver = t_driver.add_child(NonLinearSolver("solver"))
    solver.add_unknown("nonlinear.k1").add_equation("splitter.p2_out.x == 10")

    # Set BC
    t_driver.set_scenario(
        init = {"p_in.x": 1.0},
        values = {"p_in.x": "5 - 4 * exp(-t / 0.1)"},
    )
    recorder = t_driver.add_recorder(
        DataFrameRecorder(includes=["p_in.x", "p_out.x", "nonlinear.k1"]),
        period=0.1
    )
    iterativenonlinear.run_drivers()

    # Reference
    data = recorder.export_data()
    assert_almost_equal(data["p_out.x"], numpy.full(len(data), 10.0))

    # Convert to FMU
    fmu_file = to_fmu(
        iterativenonlinear,
        dest=tmp_path,
        nonlinear_solver=solver,
        locals=["nonlinear.k1"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    inputs = ("p_in.x", lambda t: 5.0 - 4.0 * numpy.exp(-t / 0.1))
    results = model.simulate(final_time=0.3, input=inputs, options={"ncp": 3})
    results = pandas.DataFrame(
        results.data_matrix.T[1:, :],
        columns=["time"] + list(model.get_model_variables()),
    )

    for column in filter(lambda c: c != "time", results.columns):
        assert results[column].values == pytest.approx(data[column].values)


def test_FMU_integration_nonlinear_local(tmp_path, iterativenonlinear):
    pyfmi = pytest.importorskip("pyfmi")

    # Init
    t_driver = iterativenonlinear.add_driver(
        RungeKutta(order=4, dt=0.1, time_interval=[0, 0.2])
    )
    solver = t_driver.add_child(NonLinearSolver("solver"))
    solver.add_child(RunSingleCase("runner"))
    solver.runner.add_unknown("nonlinear.k1").add_equation("splitter.p2_out.x == 10")

    # Set BC
    t_driver.set_scenario(
        init = {"p_in.x": 1.0},
        values = {"p_in.x": "5 - 4 * exp(-t / 0.1)"},
    )
    recorder = t_driver.add_recorder(
        DataFrameRecorder(includes=["p_in.x", "p_out.x", "nonlinear.k1"]),
        period=0.1,
    )
    iterativenonlinear.run_drivers()

    # Reference
    data = recorder.export_data()
    assert_almost_equal(data["p_out.x"], numpy.full(len(data), 10.0))

    # Convert to FMU
    fmu_file = to_fmu(
        iterativenonlinear,
        dest=tmp_path,
        nonlinear_solver=solver,
        locals=["nonlinear.k1"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    inputs = ("p_in.x", lambda t: 5.0 - 4.0 * numpy.exp(-t / 0.1))
    results = model.simulate(final_time=0.3, input=inputs, options={"ncp": 3})
    results = pandas.DataFrame(
        results.data_matrix.T[1:, :],
        columns=["time"] + list(model.get_model_variables()),
    )

    for column in filter(lambda c: c != "time", results.columns):
        assert results[column].values == pytest.approx(data[column].values)


def test_FMU_integration_vector_problem(tmp_path, vector_problem):
    pyfmi = pytest.importorskip("pyfmi")

    # Setup the system
    solver = vector_problem.add_driver(NonLinearSolver("solver"))
    

    # Init & solve
    vector_problem.x = 1.0
    vector_problem.run_drivers()

    # Convert to FMU
    fmu_file = to_fmu(
        vector_problem,
        dest=tmp_path,
        nonlinear_solver=solver,
        inputs=["x"],
        outputs=["y"],
        locals=["dummy_coef[0]"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    inputs = ("x", lambda t: 5.0 - 4.0 * numpy.exp(-t / 0.1))
    results = model.simulate(final_time=0.3, input=inputs, options={"ncp": 3})
    r = pandas.DataFrame(
        results.data_matrix.T[1:, :],
        columns=["time"] + list(model.get_model_variables()),
    )

    assert_almost_equal(r["dummy_coef[0]"].values, 0.5 * (r["x"] + r["y"]))



from cosapp.base import System, Port
from cosapp.ports.port import BasePort, BaseVariable
from cosapp.ports.enum import PortType
from cosapp.core.numerics.boundary import Unknown
from cosapp.core.numerics.residues import Residue
from cosapp.tools.views.markdown import PortMarkdownFormatter
from cosapp.utils.find_variables import make_wishlist

import json, os
import jsonschema
from pathlib import Path
from importlib import import_module
from fnmatch import fnmatchcase
from inspect import getmembers, isclass, ismodule
from typing import Set, Type, Tuple, Union, Dict, List, Any, Optional
from types import ModuleType

import logging

logger = logging.getLogger(__name__)


SearchPattern = Union[str, List[str]]
SystemConfig = Union[
    Dict[str, Dict[str, Any]],
    Dict[str, List[Dict[str, Any]]],
]


def format_kwargs(**kwargs):
    args = ", ".join(
        f"{key}={val}"
        for key, val in kwargs.items()
    )
    return f"({args})"


def extract_modules(module: ModuleType):
    """Extract all modules from `module`.
    
    Returns:
    --------
    List[module]
    """
    def recursive_search(module, modules: list):
        for _, obj in getmembers(module, ismodule):
            if obj.__name__.startswith(module.__name__) and obj not in modules:
                modules.append(obj)
                recursive_search(obj, modules)

    modules = [module]
    recursive_search(module, modules)
    return modules


def find_ports_and_systems(
    module: ModuleType,
    includes: SearchPattern = '*',
    excludes: SearchPattern = None,
) -> Tuple[Set[Type[System]], Set[Type[Port]]]:
    """Extract recursively all systems and ports from `module`

    Parameters:
    -----------
    - module [ModuleType]:
        Python module to be parsed.
    - includes [str or List[str]] (optional):
        System and port names matching these patterns will be included.
    - excludes [str or List[str]] (optional):
        System and port names matching these patterns will be excluded.
    
    Returns:
    --------
    systemSet, portSet: set[type[System]], set[type[Port]]
    """
    if not ismodule(module):
        raise TypeError('Argument is expected to be a module')

    includes = make_wishlist(includes)
    excludes = make_wishlist(excludes)
    def is_included(name: str) -> bool:
        include = False
        for pattern in includes:
            if fnmatchcase(name, pattern):
                include = True
                for pattern in excludes:
                    if fnmatchcase(name, pattern):
                        include = False
                        break
        return include
    
    def issubclass_strict(obj: type, base: type) -> bool:
        """Strict version of `issubclass`, returning `False` for base class."""
        return issubclass(obj, base) and obj is not base
    
    modules = extract_modules(module)
    systemSet, portSet = set(), set()
    for mod in modules:
        for _, obj in getmembers(mod):
            if isclass(obj) and is_included(obj.__name__):
                if issubclass_strict(obj, System):
                    systemSet.add(obj)
                if issubclass_strict(obj, Port):
                    portSet.add(obj)
    return systemSet, portSet


def get_data_from_class(
    dtype: Union[Type[System],Type[Port]],
    package_name: Optional[str] = None,
    ports: Optional[Set[Type[Port]]] = None,
    *args,
    **kwargs
):
    """Get informations from a system or a port

    Parameters:
    -----------
    - dtype [type[System | Port]]:
        System or Port class to be analyzed.
    - packageName [str] (optional):
       Custom package name.
    - ports [set[Port]] (optional):
        Used to check that all ports were found
        (in case a port is used in a children system not in the package for example).
    - *args, **kwargs:
        Additional arguments forwarded to class constructor, if required.

    Returns:
    --------
    dict[str, Any]
    """
    if ports is None:
        def register_port(port: BasePort):
            pass
    else:
        def register_port(port: BasePort):
            if isinstance(port, Port):
                ports.add(type(port))

    def get_all_port_data(portDict: Dict[str, BasePort]):
        portList = []
        for port in portDict.values():
            if len(port) > 0:
                portList.append(get_port_data(port))
                register_port(port)
        return portList or None

    sysPackage = package_name or dtype.__module__.split('.', maxsplit=1)[0]

    def get_port_data(port: BasePort, is_port_type=False):
        ptype = type(port)
        typename = ptype.__name__
        modname = ptype.__module__
        if is_port_type:
            portDict = {
                'name': typename,
            }
        else:
            portDict = {
                'name': port.name,
                'type': typename,
            }
        fixed_size_port = isinstance(port, Port)
        if fixed_size_port:
            portDict['pack'] = package_name or modname.split('.', maxsplit=1)[0]
        else:
            portDict['pack'] = sysPackage
        
        if not fixed_size_port or is_port_type:
            portDict['variables'] = get_port_var(port)
            desc = "\n".join(PortMarkdownFormatter(port).var_repr()[3:-1])
            if desc:
                portDict['desc'] = desc
        return portDict

    def get_var_data(variable: BaseVariable):
        data = {'name': variable.name}
        if (desc := variable.description):
            data['desc'] = desc
        if (unit := variable.unit):
            data['unit'] = unit
        return data

    def get_port_var(port: BasePort):
        return list(map(get_var_data, port.variables()))

    def get_shortest_import_path(modulePath: str, dtype: Type[System]) -> str:
        pathParts = modulePath.split('.')
        className = dtype.__name__

        upperModulePath = modulePath
        i = len(pathParts) -1
        while i > 0:
            currentPath = '.'.join(pathParts[:i])
            upperModule = import_module(currentPath)
            attr = getattr(upperModule, className, None)
            if attr is dtype:
                upperModulePath = currentPath
            i -= 1

        return upperModulePath

    def get_system_doc(systemClass: Type[System]) -> str:
        indent = 0
        doc = systemClass.__doc__
        strippedDoc = []
        if doc:
            for line in doc.split("\n"):
                if indent == 0:
                    strippedLine = line.lstrip()
                    if len(strippedLine) > 0:
                        indent = len(line) - len(strippedLine)
                else:
                    strippedLine = line[indent:]
                strippedDoc.append(strippedLine)
            return "\n".join(strippedDoc)
        return None

    def get_math_problem(system: System) -> Dict[str, Any]:
        """Extract mathematical problem metadata from `system`"""
        def extract_data(obj: Union[Unknown, Residue]):
            objDict = obj.to_dict()
            return { 'context': objDict['context'], 'content': objDict['name'] }
        
        def dict_to_list(objList: Union[Dict[str, Unknown], Dict[str, Residue]]):
            return list(map(extract_data, objList.values()))
    
        system.open_loops()
        problem = system.assembled_problem()
        system.close_loops()

        mathProblemDict = {}
        if (unknowns := problem.unknowns):
            mathProblemDict['unknowns'] = dict_to_list(unknowns)
            mathProblemDict['nUnknowns'] = problem.n_unknowns
        if (residues := problem.residues):
            mathProblemDict['equations'] = dict_to_list(residues)
            mathProblemDict['nEquations'] = problem.n_equations

        return mathProblemDict

    if issubclass(dtype, System):
        systemType = dtype.__name__
        systemName = kwargs.pop('__alias__', systemType)
        system = dtype('bogus', *args, **kwargs)
        desc = get_system_doc(dtype)
        inputs = get_all_port_data(system.inputs)
        outputs = get_all_port_data(system.outputs)
        mathProblemDict = get_math_problem(system)
        dtypeDict = {
            'name': systemName,
            'className': systemType,
            'pack': sysPackage,
            'mod': get_shortest_import_path(dtype.__module__, dtype),
        }
        if desc:
            dtypeDict['desc'] = desc
        if inputs:
            dtypeDict['inputs'] = inputs
        if outputs:
            dtypeDict['outputs'] = outputs
        if mathProblemDict:
            dtypeDict['mathProblem'] = mathProblemDict

    else:
        port = dtype('bogus', direction=PortType.IN, *args, **kwargs)
        dtypeDict = get_port_data(port, is_port_type=True)

    if kwargs:
        dtypeDict['kwargs'] = kwargs

    return dtypeDict


def get_data_from_module(
    module: ModuleType,
    ctor_config: SystemConfig = {},
    package_name: Optional[str] = None,
    includes: SearchPattern = '*',
    excludes: SearchPattern = None,
) -> Dict[str, Any]:
    """Extract metadata of all systems and ports found in `module`.

    Parameters:
    -----------
    - module [ModuleType]:
        Python module to be parsed.
    - ctor_config [dict[str, Any] | dict[str, list[dict[str, any]]]] (optional):
        Dictionary or list of dictionaries containing kwargs required for system/port
        construction (if any), referenced by class names (keys).
        If the dictionary contains key '__alias__', the class will be
        renamed into the associated value.
    - packageName [str] (optional):
        Custom package name.
    - includes [str or List[str]] (optional):
        System and port names matching these patterns will be included.
    - excludes [str or List[str]] (optional):
        System and port names matching these patterns will be excluded.

    Returns:
    --------
    dict[str, Any]
    """
    if package_name is None:
        package_name = module.__name__
    
    def get_all_class_data(classSet: Set[Type], ctor_config: SystemConfig, ports=None):
        result = []
        for cls in classSet:
            cls_kwargs_tab = ctor_config.get(cls.__name__, {})
            if not isinstance(cls_kwargs_tab, list):
                cls_kwargs_tab = [cls_kwargs_tab]
            for cls_kwargs in cls_kwargs_tab:
                try:
                    cls_data = get_data_from_class(cls, package_name=package_name, ports=ports, **cls_kwargs)
                except:
                    logger.info(f"Could not instantiate `{cls.__name__}`; skipped")
                    continue
                if len(cls_kwargs_tab) > 1:
                    if not cls_kwargs.pop('__alias__', None):
                        cls_data['name'] += f' {format_kwargs(**cls_kwargs)}'
                result.append(cls_data)
        return result

    systemSet, portSet = find_ports_and_systems(module, includes, excludes)
    metadata = {
        'name': package_name,
        'systems': get_all_class_data(systemSet, ctor_config, ports=portSet),
    }
    # Add port metadata *after* system metadata
    metadata['ports'] = get_all_class_data(portSet, ctor_config)

    try:
        metadata['version'] = module.__version__
    except AttributeError:
        pass

    pwd = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(pwd, "package_metadata.schema.json")) as fp:
        config_schema = json.load(fp)

    jsonschema.validate(metadata, config_schema)

    return metadata


def parse_module(
    module: ModuleType,
    ctor_config: Optional[SystemConfig] = None,
    package_name: Optional[str] = None,
    includes: Optional[SearchPattern] = None,
    excludes: Optional[SearchPattern] = None,
    path: Optional[Union[str, Path]] = None,
) -> None:
    """Creates a json file containing the metadata
    of all systems and ports found in `module`.
    
    Parameters:
    -----------
    - module [ModuleType]:
        Python module to be parsed.
    - ctor_config [dict[str, list[dict[str, any]] | dict[str, Any]]] (optional):
        Dictionary or list of dictionaries containing kwargs required for system/port
        construction (if any), referenced by class names (keys).
        If the dictionary contains key '__alias__', the class will be
        renamed into the associated value.
    - package_name [str] (optional):
        Custom package name.
    - includes [str or List[str]] (optional):
        System and port names matching these patterns will be included.
    - excludes [str or List[str]] (optional):
        System and port names matching these patterns will be excluded
        (ports used by included systems will always be included).
    - path [str | pathlib.Path] (optional):
        Optional path of output file <packageName or module.__name__>.json
        (current directory by default).

    Use pre-defined settings
    ------------------------
    Pre-defined values of optional arguments `ctor_config`, `package_name`, `includes` and `excludes`
    may be specified at module level by implementing hook function `_parse_module_config`,
    returning preset values in a dictionary of the kind {option: value}.
    Typically, `_parse_module_config` may be defined in the `__init__.py` file of the module.
    
    Examples:
    ---------
    >>> parse_module(module1)
    >>>
    >>> parse_module(
    >>>     module2,
    >>>     ctor_config = {
    >>>         'SystemA': dict(n=2, x=0.5),
    >>>         'SystemB': [
    >>>             dict(foo=0),
    >>>             dict(foo=None, __alias__='SystemB [default]'),
    >>>         ],
    >>>     },
    >>> )
    """
    settings = dict(
        ctor_config=ctor_config,
        package_name=package_name,
        includes=includes,
        excludes=excludes,
    )
    default_settings = dict(
        ctor_config={},
        includes='*',
        excludes=None,
        package_name=module.__name__,
    )
    try:
        # Check if module has a pre-defined settings
        get_module_settings = getattr(module, "_parse_module_config")
    except AttributeError:
        pass
    else:
        logger.info("Use pre-defined settings returned by `_parse_module_config`")
        default_settings.update(get_module_settings())

    # Apply user settings, if specified
    for name, value in settings.items():
        if value is None:
            settings[name] = default_settings[name]

    metadata = get_data_from_module(module, **settings)
    package_name = settings['package_name']

    try:
        version = metadata['version']
    except KeyError:
        filename = f"{package_name}.json"
    else:
        filename = f"{package_name} - {version}.json"

    if path and os.path.isdir(path):
        filename = Path(path) / filename

    with open(filename, "w") as fp:
        json.dump(metadata, fp, indent=4)

    logger.info(f"Created file {filename}")



class ContextLock:
    """Simple on/off context manager to handle locking mechanisms."""

    def __init__(self):
        self.__active = False

    def __enter__(self):
        self.__active = True

    def __exit__(self, *args, **kwargs):
        self.__active = False

    @property
    def is_active(self) -> bool:
        """bool: Is context activated?"""
        return self.__active



"""
Various small helper functions.
"""
import os
import inspect
import numpy
from numbers import Number
from typing import Any, Callable, Iterable, Type, Union, Tuple
from collections.abc import Collection


def is_numerical(value: Any) -> bool:
    """Test if a value is numerical based on its type.

    Currently type considered as numerical are:

    - int; but not bool
    - float; including numpy.inf and numpy.nan
    - complex
    - numpy.ndarray of dtype numpy.number
    - Collection convertible in a numpy array of dtype derived from numpy.number

    Parameters
    ----------
    value : Any
        Value to test

    Returns
    -------
    bool
        Is the value numerical type?
    """
    if isinstance(value, (bool, str)):
        # to avoid bool being considered as int
        # and because str is Collection
        return False
    if isinstance(value, Number):
        return True
    if isinstance(value, numpy.ndarray):
        return numpy.issubdtype(value.dtype, numpy.number)
    if isinstance(value, Collection) and len(value) > 0:
        return all(is_numerical(v) for v in value)
    return False


def is_number(value: Any) -> bool:
    """Test if a value is a Number or a 0d numerical array

    Currently type considered as numerical are:

    - int; but not bool
    - float; including numpy.inf and numpy.nan
    - complex
    - numpy.ndarray of dtype numpy.number and ndim == 0

    Parameters
    ----------
    value : Any
        Value to test

    Returns
    -------
    bool
        Is the value numerical type?
    """
    if isinstance(value, bool):
        # to avoid bool being considered as int
        return False
    if isinstance(value, Number):
        return True
    if isinstance(value, numpy.ndarray) and value.ndim == 0:
        return numpy.issubdtype(value.dtype, numpy.number)
    return False


def get_typename(dtype: Union[Type, Tuple[Type]], multiformat="({})") -> str:
        if inspect.isclass(dtype):
            return dtype.__qualname__
        return multiformat.format(", ".join(set(t.__qualname__ for t in dtype)))


def check_arg(
    arg: Any,
    argname: str,
    dtype: Union[Type, Iterable[Type]],
    value_ok: Callable[[Any], bool] = None,
    stack_shift: int = 0,
):
    """
    Utility function for argument type and value validation.
    Raises a TypeError exception if type(arg) is not in type list given by 'dtype'.
    Raises a ValueError exception if value_ok(arg) is False, where 'value_ok' is a
    boolean function defining a validity criterion.

    For example:
    >>> check_arg(-0.12, 'my_var', float)
    does not raise any exception, as -0.12 is a float

    >>> check_arg(-0.12, 'my_var', (int, str))
    raises TypeError, as first argument is neither an int, not a str

    >>> check_arg(-0.12, 'my_var', float, value_ok = lambda x: x > 0)
    raises ValueError, as first argument is not strictly positive
    """
    def get_caller():
        level = 3 + max(0, stack_shift)
        stack = inspect.stack()
        return stack[level] if len(stack) > level else stack[-1]

    def get_context(caller):
        try:
            context = caller.code_context[0]
        except:
            context = ""
        return os.path.basename(caller.filename), caller.lineno, context

    # Check type
    if not isinstance(arg, dtype):
        valid = get_typename(dtype, multiformat="one of ({})")
        caller = get_caller()
        raise TypeError("argument '{}' should be {}; got {} {!r}\nIn {}, line #{}: \n{}".format(
            argname, valid, type(arg).__qualname__, arg, *get_context(caller)))
    # Check value
    if isinstance(value_ok, Callable) and not value_ok(arg):
        caller = get_caller()
        raise ValueError("argument {!r} was given invalid value {!r}\nIn {}, line #{}: \n{}".format(
            argname, arg, *get_context(caller)))


def partition(iterable: Iterable[Any], predicate: Callable[[Any], bool]):
    """Partition a collection into two lists, using filter function `predicate`.

    Parameters:
    -----------
    - iterable: Iterable[Any]
        Iterable collection of elements confronted to predicate.
    - predicate: Callable[[Any], bool]
        Boolean function used to partition the collection.

    Returns:
    --------
    yays, nays: tuple[list, list]
        Lists containing the elements for which predicate is `True` (yays) and `False` (nays).
    """
    yays, nays = [], []

    for element in iterable:
        if predicate(element):
            yays.append(element)
        else:
            nays.append(element)
    
    return yays, nays



"""Test problems with unknown aliased by pulling"""
import pytest
import logging, re
from numpy import sqrt, cbrt

from cosapp.systems import System
from cosapp.drivers import NonLinearSolver, RunSingleCase


class Quadratic(System):
    """System computing y = 1 + k * x**2"""
    def setup(self):
        self.add_inward('k', 0.8)
        self.add_inward('x', 1.0)
        self.add_outward('y', 0.0)
    
    def compute(self):
        self.y = 1 + self.k * self.x**2


class NestedQuad(System):
    """System combining two Quadratic subsystems, with pulling"""
    def setup(self):
        foo = self.add_child(Quadratic('foo'), pulling={'k': 'c'})
        bar = self.add_child(Quadratic('bar'))

        self.connect(foo, bar, {'y': 'k'})


def test_design_unknown_aliasing_1(caplog):
    """Design test using an unknown aliased by a pulling.
    System is a `NestedQuad` instance.

    Case 1: Use pulled input variable `foo.k` as unknown.

    Successful; log simply informs that unknown was substituted.
    """
    top = NestedQuad('top')
    solver = top.add_driver(NonLinearSolver('solver', tol=1e-9))
    solver.add_child(RunSingleCase('runner'))
    solver.runner.add_equation('foo.y == 7').add_unknown('foo.k')
    solver.runner.set_values({
        'foo.x': 2,
        'bar.x': 0.1,
    })
    caplog.clear()
    with caplog.at_level(logging.INFO):
        top.run_drivers()
    
    assert len(caplog.records) > 0
    assert re.match(
        "Replace unknown 'foo.k' by 'c'",
        caplog.records[0].message
    )
    assert top.foo.k == pytest.approx(1.5)
    assert top.foo.y == pytest.approx(7)
    assert top.bar.k == top.foo.y
    assert top.foo.k == top.c


def test_design_unknown_aliasing_2(caplog):
    """Design test using an unknown aliased by a pulling.
    System is a `NestedQuad` instance.

    Case 2:
        Driver is attached to subsystem `foo`.
        Unknown is pulled inward `foo.k`.
    
    Works, but issues a warning because unknown is aliased
    by a higher-level, free input.
    """
    top = NestedQuad('top')
    top.c = 0.1
    top.run_once()
    assert top.c == 0.1
    assert top.foo.k == top.c

    solver = top.foo.add_driver(NonLinearSolver('solver', tol=1e-9))
    solver.add_child(RunSingleCase('runner'))
    solver.runner.add_unknown('k').add_equation('y == 7')
    solver.runner.set_values({'x': 2})

    caplog.clear()
    with caplog.at_level(logging.WARNING):
        top.run_drivers()
    
    assert len(caplog.records) > 0
    assert any(
        re.match(
            "Unknown 'foo.k' is aliased by 'top.c', defined outside the context of 'foo'",
            record.message
        )
        for record in caplog.records
    )
    assert top.foo.k == pytest.approx(1.5)
    assert top.c == 0.1  # unchanged top value


def test_design_unknown_aliasing_3(caplog):
    """Same as `test_design_unknown_aliasing_1`, except `NestedQuad` system
    is now a sub-system in which inward `c` is connected to an output.
    """
    top = System('top')
    sub1 = top.add_child(Quadratic('sub1'))
    sub2 = top.add_child(NestedQuad('sub2'))
    top.connect(sub1.outwards, sub2.inwards, {'y': 'c'})  # sub2.c is no longer free

    solver = top.add_driver(NonLinearSolver('solver'))
    # Use input `bar.k` connected to an output as unknown:
    solver.add_equation('sub2.foo.y == 7').add_unknown('sub2.bar.k')

    caplog.clear()
    with pytest.raises(ArithmeticError, match="numbers of params \[0\] and residues \[1\]"):
        with caplog.at_level(logging.WARNING):
            top.run_drivers()
    
    assert len(caplog.records) > 0
    assert re.match(
        "Skip connected unknown 'sub2.bar.k'",
        caplog.records[0].message
    )


def test_design_unknown_aliasing_4(caplog):
    """Same as `test_design_unknown_aliasing_1`, with
    one additional hierarchical level.
    """
    top = System('top')
    sub = top.add_child(NestedQuad('sub'))
    solver = top.add_driver(NonLinearSolver('solver', tol=1e-9))
    solver.add_equation('sub.foo.y == 7').add_unknown('sub.foo.k')

    sub.foo.x = 2.0
    sub.bar.x = 0.1
    caplog.clear()
    with caplog.at_level(logging.INFO):
        top.run_drivers()
    
    assert sub.foo.k == pytest.approx(1.5)
    assert sub.foo.y == pytest.approx(7)
    assert sub.bar.k == sub.foo.y
    assert sub.foo.k == sub.c
    assert len(caplog.records) > 0
    assert re.match(
        "Replace unknown 'sub.foo.k' by 'sub.c'",
        caplog.records[0].message
    )


def test_design_connected_unknown_1(caplog):
    """Tests that unknowns connected to an output variable are discarded"""
    top = NestedQuad('top')
    solver = top.add_driver(NonLinearSolver('solver'))
    # Use input `bar.k` connected to an output as unknown:
    solver.add_equation('bar.y == 7').add_unknown('bar.k')

    caplog.clear()
    with pytest.raises(ArithmeticError, match="numbers of params \[0\] and residues \[1\]"):
        with caplog.at_level(logging.WARNING):
            top.run_drivers()
    
    assert len(caplog.records) > 0
    assert re.match(
        "Skip connected unknown 'bar.k'",
        caplog.records[0].message
    )


def test_design_connected_unknown_2(caplog):
    """Same as `test_design_connected_unknown_1`,
    except driver is attached to subsystem `bar`.
    
    Fails because unknown is connected.
    """
    top = NestedQuad('top')
    # Attach driver to sub-system `bar`
    solver = top.bar.add_driver(NonLinearSolver('solver', tol=1e-9))
    solver.add_unknown('k').add_equation('y == 7')

    caplog.clear()
    with pytest.raises(ArithmeticError, match="numbers of params \[0\] and residues \[1\]"):
        with caplog.at_level(logging.WARNING):
            top.run_drivers()
    
    assert len(caplog.records) > 0
    assert re.match(
        "Skip connected unknown 'k'",
        caplog.records[0].message
    )


def test_design_offdesign_aliasing_1(caplog):
    """Design test involving off-design and design unknowns
    aliased by the same variable.

    Case 1: free unknown is off-design
    """
    class TopSystem(NestedQuad):
        """NestedQuad with off-design problem"""
        def setup(self):
            super().setup()
            # Use parent (free) variable 'c' as off-design unknown
            self.add_unknown('c').add_equation('c == 0')
    
    top = TopSystem('top')
    solver = top.add_driver(NonLinearSolver('solver', tol=1e-9))
    # Use pulled variable `foo.k` as design unknown:
    solver.add_equation('foo.y == 7').add_unknown('foo.k')

    caplog.clear()
    with pytest.raises(ValueError, match="'c' is defined as design and off-design unknown"):
        with caplog.at_level(logging.INFO):
            top.run_drivers()
    
    assert len(caplog.records) > 0
    assert any(
        re.match(
            "Replace unknown 'foo.k' by 'c'",
            record.message
        )
        for record in caplog.records
    )


def test_design_offdesign_aliasing_2(caplog):
    """Design test involving off-design and design unknowns
    aliased by the same variable.

    Case 2: free unknown is declared in design problem
    """
    class TopSystem(NestedQuad):
        """NestedQuad with off-design problem"""
        def setup(self):
            super().setup()
            # Use pulled variable `foo.k` as off-design unknown
            self.add_unknown('foo.k').add_equation('c == 0')
    
    top = TopSystem('top')
    solver = top.add_driver(NonLinearSolver('solver', tol=1e-9))
    # Use parent variable 'c' as design unknown
    solver.add_equation('foo.y == 7').add_unknown('c')

    caplog.clear()
    with pytest.raises(ValueError, match="'c' is defined as design and off-design unknown"):
        with caplog.at_level(logging.INFO):
            top.run_drivers()
    
    assert len(caplog.records) > 0
    assert any(
        re.match(
            "Replace unknown 'foo.k' by 'c'",
            record.message
        )
        for record in caplog.records
    )


def test_design_offdesign_aliasing_3(caplog):
    """Design test involving off-design and design unknowns
    aliased by the same variable.

    Case 3: free unknown is off-design; design unknown defined at solver level
    """
    class TopSystem(NestedQuad):
        """NestedQuad with off-design problem"""
        def setup(self):
            super().setup()
            # Use parent (free) variable 'c' as off-design unknown
            self.add_unknown('c').add_equation('c == 0')
    
    top = TopSystem('top')
    solver = top.add_driver(NonLinearSolver('solver', tol=1e-9))
    # Use pulled variable `foo.k` as design unknown:
    solver.add_equation('foo.y == 7').add_unknown('foo.k')

    caplog.clear()
    with pytest.raises(ValueError, match="'c' is defined as design and off-design unknown"):
        with caplog.at_level(logging.INFO):
            top.run_drivers()
            # print(solver.problem)
    
    assert len(caplog.records) > 0
    assert any(
        re.match(
            "Replace unknown 'foo.k' by 'c'",
            record.message
        )
        for record in caplog.records
    )


def test_connected_unknown_changing_conf(caplog):
    """Test with an originally free input variable,
    later connected to an output after a first design problem.
    """
    class DoubleQuad(System):
        def setup(self):
            self.add_child(Quadratic('foo'))
            self.add_child(Quadratic('bar'))

    top = DoubleQuad('top')
    foo, bar = top.foo, top.bar
    solver = top.add_driver(NonLinearSolver('solver', tol=1e-9, max_iter=100))

    # First problem: `bar.x` is unknown
    solver.add_unknown('bar.x').add_equation('bar.y == 5.5')

    foo.k = 1
    bar.k = 2
    top.run_drivers()
    assert bar.y == pytest.approx(5.5)
    assert bar.x == pytest.approx(1.5)
    assert set(solver.problem.unknowns) == {'bar.x'}
    assert set(solver.problem.residues) == {'bar.y == 5.5'}

    # Add connector foo.y -> bar.x
    # Unknown `bar.x` is no longer free
    top.connect(foo, bar, {'y': 'x'})

    caplog.clear()
    with pytest.raises(ArithmeticError, match="numbers of params \[0\] and residues \[1\]"):
        with caplog.at_level(logging.WARNING):
            top.run_drivers()
    
    assert len(caplog.records) > 0
    assert re.match(
        "Skip connected unknown 'bar.x'",
        caplog.records[0].message
    )

    # Declare `foo.x` as unknown - problem is balanced again
    solver.add_unknown('foo.x')

    top.run_drivers()
    assert foo.x == pytest.approx(sqrt(0.5))
    assert bar.y == pytest.approx(5.5)
    assert set(solver.problem.unknowns) == {'foo.x'}
    assert set(solver.problem.residues) == {'bar.y == 5.5'}

    # Introduce cyclic dependency with new sub-system
    class CubicRoot(System):
        def setup(self):
            self.add_inward('x', 1.0)
            self.add_outward('y', 1.0)
        
        def compute(self) -> None:
            self.y = cbrt(self.x)

    coupling = CubicRoot('coupling')
    top.add_child(coupling)
    # Make `foo.k` a function of `bar.y`
    # with:
    #     bar.y -> coupling.x
    #     coupling.y -> foo.k
    top.connect(bar, coupling, {'y': 'x'})
    top.connect(coupling, foo, {'y': 'k'})

    assert top.assembled_problem().is_empty()
    top.open_loops()
    problem = top.assembled_problem()
    assert problem.shape == (1, 1)
    assert set(problem.unknowns) == {'foo.k'}
    assert set(problem.residues) == {'foo.k == coupling.y (loop)'}
    top.close_loops()

    # Cyclic dependency induces a change in system structure,
    # which must be captured at driver execution
    top.run_drivers()
    assert bar.y == pytest.approx(5.5)
    assert foo.k == pytest.approx(cbrt(5.5))
    assert foo.x == pytest.approx(0.5322200367)
    assert set(solver.problem.unknowns) == {
        'foo.x',
        'foo.k',
    }
    assert set(solver.problem.residues) == {
        'bar.y == 5.5',
        'foo.k == coupling.y (loop)',
    }



"""Build a D3js view of a System."""

import os
from typing import Dict
from cosapp.systems import System
from cosapp.tools import templates
from .baseRenderer import BaseRenderer


class D3JsRenderer(BaseRenderer):
    """
    Utility class to export a system as HTML using D3.JS library.

    Parameters
    ----------
    system : System
        System to export

    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    
    __globals = None

    def __init__(self, system: System, embeddable=False):
        super().__init__(system, embeddable)

    @staticmethod
    def get_level(syst) -> int:
        """
        Return the number of child levels of input system
        """
        if len(syst.children) == 0:
            return 1
        else:
            return 1 + max(map(D3JsRenderer.get_level, syst.children.values()))

    def get_data(self) -> Dict:
        """
        Convert  `self.system` into a dictionary with 4 keys: `name`, `full_name`, `children` and `size`.

        - `name` key is a string with the system name.
        - `full_name` key is a string with Module contextual name.
        - `children` key is a list containing dictionaries of same format for children system .
        - `size` key is the display size of system in D3 plot.

        Returns
        -------
        Dict
            Dictionary containing elements to draw the `System` using D3JS library
        """

        def build_d3_json(
            system: System, parent_name: str = None, nlevels: int = 0
        ) -> Dict[str, Dict]:
            """
            Builds the dictionary of system structure for D3 plot.

            Parameters
            ----------
            system : System
                Current input system

            parent_name : str, optional
                Name of parent system, default None

            nlevels : int, optional
                Number of system levels, default 0

            Returns
            -------
            Dict[str, Dict]
                The structure of input system in D3 format
            """
            if parent_name is None:
                full_name = system.name
            else:
                full_name = f"{parent_name}.{system.name}"

            if nlevels == 0:  # Number of System levels
                nlevels = D3JsRenderer.get_level(system)

            level = len(full_name.split("."))

            rtn = {"name": system.name, "full_name": full_name}
            if system.children:
                rtn["children"] = [
                    build_d3_json(child, parent_name=full_name, nlevels=nlevels)
                    for child in system.children.values()
                ]
            else:
                rtn["size"] = (nlevels / level) ** 5

            return rtn

        return build_d3_json(self.system)

    @classmethod
    def html_resources(cls) -> Dict[str, str]:
        """
        Return a dictionary of three keys.

        - `d3_js` key holds the content of D3.js library.
        - `draw_js` key holds the content of the draw function implemented in `d3_draw.js`.
        - `d3_styles` key holds the css style of D3 plot.

        Returns
        -------
        Dict[str, Dict]
            The necessary resources to render Jinja template
        """
        ressource_dir = os.path.dirname(os.path.abspath(templates.__file__))
        libs_dir = os.path.join(ressource_dir, "lib")
        src_dir = os.path.join(ressource_dir, "src")
        # Grab the resources
        with open(os.path.join(libs_dir, "d3.min.js"), "r", encoding="utf-8") as f:
            d3_js = f.read()
        with open(os.path.join(src_dir, "d3_draw.js"), "r", encoding="utf-8") as f:
            draw_js = f.read()
        with open(os.path.join(src_dir, "d3_styles.css"), "r", encoding="utf-8") as f:
            d3_styles = f.read()

        return {"d3_js": d3_js, "draw_js": draw_js, "d3_styles": d3_styles}

    @classmethod
    def html_template(cls) -> "Template":
        """
        Return the Jinja template used to create HTML file.

        Returns
        -------
        Template

        """
        try:
            from jinja2 import Environment, PackageLoader
        except ImportError:
            raise ImportError("jinja2 needs to be installed to export to d3 JS.")

        env = Environment(
            loader=PackageLoader("cosapp.tools", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = env.get_template("system_d3.html")

        return template

    def html_content(self) -> str:
        """
        Return D3 HTML content of renderer's system as a character string.

        Returns
        -------
        str

        """
        common = self.get_globals()
        template = common["template"]
        html_begin_tags = common["html_begin_tags"] + "\n<div class='flexdiv'>"
        html_end_tags = common["html_end_tags"]
        d3_js = common["d3_js"]
        draw_js = common["draw_js"]
        d3_styles = common["d3_styles"]

        elements = {
            "title": "my viewer",
            "model_data": f"var modelData = {self.get_data()!s}",
            "d3_styles": d3_styles,
            "d3JS": d3_js,
            "draw": draw_js,
        }

        rendered_html = template.render(elements)
        if not self.embeddable:
            html_begin_tags = html_begin_tags.replace("{{title}}", elements["title"])
            rendered_html = html_begin_tags + rendered_html + html_end_tags

        return rendered_html
    
    @classmethod
    def get_globals(cls):
        cg = cls.__globals
        if cg is None:
            cg = super().get_globals()
        return cg


def d3_html(system: System, embeddable=False) -> str:
    """Return D3.JS HTML content of `system`.

    Parameters
    ----------
    system : System
        System to export

    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    renderer = D3JsRenderer(system, embeddable)
    return renderer.html_content()


def to_d3(system: System, show: bool = True, size=435) -> None:
    """Return the representation of this system in HTML format.

    Returns
    -------
    str
        HTML formatted representation
    """

    filename = f"{system.name}_d3.html"
    rendered_html = d3_html(system, embeddable=False)
    with open(filename, mode="w", encoding="utf-8") as f:
        f.write(rendered_html)

    if show:
        try:
            from IPython.display import IFrame
        except ImportError:
            raise ImportError("IPython needs to be installed to display the d3 figure.")
        return IFrame(filename, "99.9%", f"{size}px")
    else:
        return None
